from fluid_sbom.internal.package_information.api_interface import (
    make_get,
)


def get_package_versions_html(
    name: str, distro_version: str | None = None, arch: str | None = None
) -> str | None:
    # Create a mapping for arch equivalence
    arch_mapping = {
        "x86": "x86",
        "x86_64": "x86_64",
        "amd64": "x86_64",
        "i386": "x86",
        "i686": "x86",
        "arm": "arm",
        "armhf": "arm",
        "armv7": "arm",
        "arm64": "aarch64",
        "aarch64": "aarch64",
        "ppc64le": "ppc64le",
        "s390x": "s390x",
    }

    # Use the mapping to get the equivalent arch
    arch = arch_mapping.get(arch.lower(), arch) if arch else None

    if arch is None:
        return None

    # Convert distro_version format if necessary
    if distro_version:
        parts = distro_version.split(".")
        if len(parts) == 3:
            distro_version = f"v{parts[0]}.{parts[1]}"

    response = make_get(
        "https://pkgs.alpinelinux.org/packages",
        params={
            "name": name,
            "branch": distro_version or "edge",
            "repo": "",
            "arch": arch or [],
            "maintainer": "",
        },
        content=True,
    )
    return response



import abc
from typing import (
    Generator,
    Type,
)


class RpmDBInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(
        cls: Type["RpmDBInterface"], subclass: object
    ) -> bool:
        return (
            hasattr(subclass, "read")
            and callable(subclass.read)
            or NotImplemented
        )

    @abc.abstractmethod
    def read(
        self,
    ) -> Generator[bytes, None, None]:
        """Read entry bytes"""
        raise NotImplementedError



from cpe import (
    CPE,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Advisory,
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from typing import (
    cast,
)


def _get_target_software(cpe: str) -> list[str]:
    return cast(
        list[str],
        CPE(cpe).get_target_software(),  # type: ignore[misc]
    )


def _matches_cpe(package: Package, advisory: Advisory) -> bool:
    mapping: dict[Language, list[str]] = {
        Language.DART: ["dart"],
        Language.DOTNET: [".net", "asp.net"],
        Language.GO: ["go", "golang"],
        Language.JAVA: ["java"],
        Language.JAVASCRIPT: ["javascript", "node.js", "nodejs"],
        Language.PHP: ["php"],
        Language.PYTHON: ["pypi", "python"],
        Language.RUBY: ["rails", "ruby", "ruby_on_rails"],
        Language.RUST: ["rust"],
        Language.SWIFT: ["swift"],
    }

    return package.language in mapping and any(
        target_software in mapping[package.language]
        for cpe in advisory.cpes
        for target_software in _get_target_software(cpe)
    )


def matches_namespace(package: Package, advisory: Advisory) -> bool:
    namespace_type = advisory.namespace.split(":")[1]
    if namespace_type == "language":
        return package.language.value in {
            "dotnet",
            "go",
            "java",
            "javascript",
            "php",
            "python",
            "ruby",
            "rust",
            "swift",
        }

    if namespace_type == "distro":
        return package.type in {
            PackageType.DebPkg,
            PackageType.ApkPkg,
            PackageType.RpmPkg,
        }
    if namespace_type == "cpe":
        return _matches_cpe(package, advisory)

    return False



from contextlib import (
    suppress,
)
from fluid_sbom.internal.collection.types import (
    FileCoordinate,
    IndexedDict,
    IndexedList,
    Position,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedChildrenLength,
    UnexpectedNode,
    UnexpectedNodeType,
)
import os
import re
from tree_sitter import (
    Language,
    Node,
    Parser,
)
from typing import (
    Any,
    cast,
)


def _generate_position(node: Node) -> Position:
    return Position(
        start=FileCoordinate(
            line=node.start_point[0] + 1, column=node.start_point[1] + 1
        ),
        end=FileCoordinate(
            line=node.end_point[0] + 1, column=node.end_point[1] + 1
        ),
    )


def _handle_block_mapping_node(node: Node) -> tuple[Node, IndexedDict]:
    data = IndexedDict(node)
    for child in node.children:
        if child.type != "block_mapping_pair":
            continue
        childs = tuple(x for x in child.children if x.type != "comment")
        if len(childs) != 3:
            childs = (*childs, None)  # type: ignore
            if childs[1].type != ":":
                raise UnexpectedChildrenLength(node, 3)
        key_up_node, _, value_up_node = childs
        key_node, key_value = handle_node(key_up_node)
        if value_up_node is None:
            data[  # type: ignore
                (
                    key_value,
                    _generate_position(key_node),
                )
            ] = (
                None,
                None,
            )
            continue
        value_node, value_value = handle_node(value_up_node)
        data[(key_value, _generate_position(key_node))] = (
            value_value,
            _generate_position(value_node),
        )
    return node, data


def _handle_flow_mapping_node(node: Node) -> tuple[Node, IndexedDict]:
    pair_nodes = [x for x in node.children if x.type == "flow_pair"]
    data = IndexedDict(node)
    for pair_node in pair_nodes:
        if len(pair_node.children) != 3:
            raise ValueError("Unexpected node")
        key_up_node, _, value_up_node = pair_node.children
        key_node, key_value = handle_node(key_up_node)
        value_node, value_value = handle_node(value_up_node)
        data[(key_value, _generate_position(key_node))] = (
            value_value,
            _generate_position(value_node),
        )
    return node, data


def _handle_boolean_scalar_node(node: Node) -> tuple[Node, bool]:
    return node, node.text.decode("utf-8").lower() == "true"


def _handle_block_sequence_node(node: Node) -> tuple[Node, IndexedList]:
    data = IndexedList(node)
    for child in (x for x in node.children if x.type != "comment"):
        if child.type != "block_sequence_item":
            raise UnexpectedNodeType(child.type, "block_sequence_item")
        resolved_item = handle_node(child)
        data.append((resolved_item[1], _generate_position(resolved_item[0])))
    return node, data


def _handle_block_sequence_item(node: Node) -> tuple[Node, Any]:
    if len(node.children) != 2 or node.children[0].type != "-":
        raise UnexpectedNodeType(node)
    return handle_node(node.children[1])


def _handle_integer_scalar_node(node: Node) -> tuple[Node, int]:
    decode_str = node.text.decode("utf-8")
    with suppress(ValueError):
        return node, int(decode_str)
    with suppress(ValueError):
        return node, int(decode_str, 16)
    raise ValueError(f"Invalid integer value: {decode_str}")


def _handle_flow_sequence_node(node: Node) -> tuple[Node, IndexedList]:
    data = IndexedList(node)
    for child in [x for x in node.children if x.type not in ("[", "]", ",")]:
        if child.type != "flow_node":
            raise UnexpectedNodeType(child.type, "flow_node")
        resolved_node, resolved_item = handle_node(child)
        data.append((resolved_item, _generate_position(resolved_node)))
    return node, data


def _handle_float_scalar_node(node: Node) -> tuple[Node, float]:
    decoded_str = node.text.decode("utf-8")
    with suppress(ValueError):
        return node, float(decoded_str)

    with suppress(ValueError):
        return node, float(decoded_str.replace(".", "").lower())

    raise ValueError(f"Invalid float value: {decoded_str}")


def _handle_block_scalar(node: Node) -> tuple[Node, str]:
    decoded_str = node.text.decode("utf-8")
    value = ""
    if match := re.search(r"^>(\d+)", decoded_str):
        indent_spaces = int(match.group(1))
        decoded_str = re.sub(r"^>\d+", "", decoded_str).strip()
    else:
        indent_spaces = 1  # Default to no indent removal if no marker is found

    if decoded_str.startswith(">"):
        value = (" " * indent_spaces).join(
            decoded_str.lstrip(">+-").strip().split()
        )
    elif decoded_str.startswith("|"):
        normalized_str = decoded_str.replace("\xa0", " ").lstrip("|+-")
        value = "\n".join(line.strip() for line in normalized_str.split("\n"))
        value = value.replace("\n", "", 1)
    return node, value


def handle_node(node: Node) -> tuple[Node, Any]:
    value: tuple[Node, Any] | None = None
    match node.type:
        case "block_node":
            if len(node.children) > 1:
                raise UnexpectedChildrenLength(node.type, 1)
            value = handle_node(node.children[0])
        case "block_mapping":
            value = _handle_block_mapping_node(node)
        case "flow_node" | "plain_scalar":
            if len(node.children) > 1:
                raise UnexpectedChildrenLength(node.type, 1)
            value = handle_node(node.children[0])
        case "string_scalar":
            value = node, node.text.decode("utf-8")
        case "single_quote_scalar" | "double_quote_scalar":
            value = node, node.text.decode("utf-8").strip("'\"")
        case "float_scalar":
            value = _handle_float_scalar_node(node)
        case "flow_mapping":
            value = _handle_flow_mapping_node(node)
        case "boolean_scalar":
            value = _handle_boolean_scalar_node(node)
        case "null_scalar":
            value = node, None
        case "integer_scalar":
            value = _handle_integer_scalar_node(node)
        case "block_sequence_item":
            value = _handle_block_sequence_item(node)
        case "block_sequence":
            value = _handle_block_sequence_node(node)
        case "flow_sequence":
            value = _handle_flow_sequence_node(node)
        case "block_scalar":
            value = _handle_block_scalar(node)
        case _:
            raise UnexpectedNode(node.type)
    return value


def parse_yaml_with_tree_sitter(
    content: str,
) -> IndexedDict | IndexedList | None:
    parser = Parser()
    parser.set_language(
        Language(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "yaml.so",
            ),
            "yaml",
        )
    )

    result = parser.parse(content.encode("utf-8"))
    documents = [x for x in result.root_node.children if x.type == "document"]
    if len(documents) != 1:
        return None
    block_node = next(
        (x for x in documents[0].children if x.type == "block_node"), None
    )
    if not block_node:
        return None
    _, value = handle_node(block_node)
    return cast(IndexedDict | IndexedList, value)



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.javascript.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.npm import (
    NpmPackage,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
from pydantic import (
    ValidationError,
)
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def _create_package(
    package_json: IndexedDict,
    reader: LocationReadCloser,
    package_name: str,
    specifier: str,
    is_dev: bool,
) -> Package | None:
    current_location = deepcopy(reader.location)
    if current_location.coordinates:
        dependencies_key = "devDependencies" if is_dev else "dependencies"
        current_location.coordinates.line = (
            package_json[dependencies_key]
            .get_key_position(package_name)
            .start.line
        )
    try:
        return Package(
            name=package_name,
            version=specifier,
            type=PackageType.NpmPkg,
            language=Language.JAVASCRIPT,
            licenses=[],
            locations=[current_location],
            p_url=package_url(package_name, specifier),
            metadata=NpmPackage(
                name=package_name,
                version=specifier,
                is_dev=is_dev,
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": current_location.path(),
                }
            },
        )
        return None


def parse_package_json(
    _: Resolver | None, __: Environment | None, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    try:
        package_json: IndexedDict = cast(
            IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
        )
    except (UnexpectedNode, ValueError):
        LOGGER.warning(
            "Failed to parse package.json",
            extra={
                "extra": {"location": reader.location},
            },
        )
        return [], []

    packages = []

    for package_name, specifier in package_json.get(
        "dependencies", {}
    ).items():
        package = _create_package(
            package_json, reader, package_name, specifier, is_dev=False
        )
        if package:
            packages.append(package)

    for package_name, specifier in package_json.get(
        "devDependencies", {}
    ).items():
        package = _create_package(
            package_json, reader, package_name, specifier, is_dev=True
        )
        if package:
            packages.append(package)

    return packages, []



from fluid_sbom.internal.package_information.dotnet import (
    get_nuget_package,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
from typing import (
    Any,
)


def _set_health_metadata_artifact(
    package: Package,
    current_package: dict[str, Any] | None,
) -> Package:
    if current_package:
        # {LOWER_ID}/{LOWER_VERSION}/{LOWER_ID}.{LOWER_VERSION}.nupkg
        lower_id = current_package["id"].lower()
        lower_version = current_package["version"].lower()
        if package.health_metadata:
            digest_value = current_package.get("packageHash") or None
            algorithm = current_package.get("packageHashAlgorithm") or None
            package.health_metadata.artifact = Artifact(
                url=(
                    f"https://api.nuget.org/v3-flatcontainer/{lower_id}"
                    f"/{lower_version}/{lower_id}.{lower_version}.nupkg"
                ),
                integrity=Digest(
                    algorithm=algorithm if digest_value else None,
                    value=digest_value,
                ),
            )
    return package


def complete_package(package: Package) -> Package:
    current_package = get_nuget_package(package.name, package.version)
    nuget_package = get_nuget_package(package.name)
    if not nuget_package:
        return package
    try:
        package.health_metadata = HealthMetadata(
            latest_version=nuget_package["version"],
            latest_version_created_at=nuget_package["published"],
        )
    except IndexError:
        return package
    package = _set_health_metadata_artifact(package, current_package)
    if package.health_metadata:
        package.health_metadata.authors = nuget_package["authors"]
    if not package.licenses and nuget_package["licenseExpression"]:
        package.licenses = [nuget_package["licenseExpression"]]

    return package



from fluid_sbom.config.logger import (
    LOGGER,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.constants import (
    HASH_OF_INDEX_PAGE_TYPE,
    HASH_PAGE_TYPE,
    HASH_UNSORTED_PAGE_TYPE,
    VALID_PAGE_SIZES,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.hash_metadata_page import (
    HashMetadataPage,
    parse_hash_metadata_page,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.hash_page import (
    hash_page_value_content,
    hash_page_value_indexes,
    parse_hash_page,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.errors import (
    InvalidDBFormat,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)
import io
from typing import (
    Generator,
)


class BerkeleyDB(RpmDBInterface):
    def __init__(self, file_path: str):
        self.file_path = file_path
        self.file = open(file_path, "rb")  # pylint:disable=consider-using-with
        self.hash_metadata = self._read_metadata()

    def _read_metadata(self) -> HashMetadataPage:
        # Read the first 512 bytes to parse metadata
        self.file.seek(0)
        metadata_buff = self.file.read(512)
        if len(metadata_buff) < 512:
            raise InvalidDBFormat("Failed to read metadata: insufficient data")

        # Parse the hash metadata page
        hash_metadata = parse_hash_metadata_page(metadata_buff)

        # Validate the page size
        if (
            hash_metadata.hash_metadata.generic_metadata_page.page_size
            not in VALID_PAGE_SIZES
        ):
            raise InvalidDBFormat(
                "Unexpected page size: "
                + str(
                    hash_metadata.hash_metadata.generic_metadata_page.page_size
                )
            )

        return hash_metadata

    def close(self) -> None:
        self.file.close()

    def read(self) -> Generator[bytes, None, None]:
        page_size = (
            self.hash_metadata.hash_metadata.generic_metadata_page.page_size
        )
        swapped = self.hash_metadata.swapped
        last_page_no = (
            self.hash_metadata.hash_metadata.generic_metadata_page.last_page_no
        )

        self.file.seek(0, io.SEEK_SET)

        for _ in range(last_page_no + 1):
            page_data = self._read_page(page_size)
            if not page_data:
                return

            hash_page_header = parse_hash_page(page_data, swapped)
            if hash_page_header.page_type not in (
                HASH_UNSORTED_PAGE_TYPE,
                HASH_PAGE_TYPE,
            ):
                continue

            hash_page_indexes = hash_page_value_indexes(
                page_data, hash_page_header.num_entries, swapped
            )

            for hash_page_index in hash_page_indexes:
                if page_data[hash_page_index] != HASH_OF_INDEX_PAGE_TYPE:
                    continue

                current_file_position = self.file.tell()
                try:
                    value_content = hash_page_value_content(
                        self.file,
                        page_data,
                        hash_page_index,
                        page_size,
                        swapped,
                    )
                    yield value_content
                except ValueError:
                    return
                finally:
                    self.file.seek(current_file_position, io.SEEK_SET)

    def _read_page(self, page_size: int) -> bytes:
        page_data = self.file.read(page_size)
        if len(page_data) != page_size:
            LOGGER.exception(
                "Failed to read page",
                extra={"extra": {"location": self.file_path}},
            )
            return b""
        return page_data



from datetime import (
    datetime,
    UTC,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.config.config import (
    SbomConfig,
)
from fluid_sbom.format.common import (
    set_namespace_version,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.exceptions import (
    SPDXValidationError,
)
import logging
import os
import re
from spdx_tools.spdx.model.actor import (
    Actor,
    ActorType,
)
from spdx_tools.spdx.model.document import (
    CreationInfo,
    Document,
)
from spdx_tools.spdx.model.package import (
    ExternalPackageRef,
    ExternalPackageRefCategory,
    Package as SPDX_Package,
    PackagePurpose,
)
from spdx_tools.spdx.model.relationship import (
    Relationship as SPDXRelationship,
    RelationshipType as SPDXRelationshipType,
)
from spdx_tools.spdx.model.spdx_no_assertion import (
    SpdxNoAssertion,
)
from spdx_tools.spdx.validation.document_validator import (
    validate_full_spdx_document,
)
from spdx_tools.spdx.validation.validation_message import (
    ValidationMessage,
)
from spdx_tools.spdx.writer.write_anything import (
    write_file,
)
from typing import (
    cast,
)
from urllib.parse import (
    urlunparse,
)
import uuid

LOGGER = logging.getLogger(__name__)

NOASSERTION = SpdxNoAssertion()


def clear_name(text: str) -> str:
    return re.sub(r"[^a-zA-Z0-9.-]", "-", text)


def get_spdx_id(package: Package) -> str:
    name = clear_name(package.name)
    pkg_platform = clear_name(package.type.value.lower())
    pkg_id = package.id_
    return f"SPDXRef-Package-{pkg_platform}-{name}-{pkg_id}"


def document_namespace(working_dir: str) -> str:
    input_type = "unknown-source-type"

    if os.path.isfile(working_dir):
        input_type = "file"
    elif os.path.isdir(working_dir):
        input_type = "dir"

    unique_id = uuid.uuid4()
    identifier = os.path.join(input_type, str(unique_id))
    if working_dir != ".":
        identifier = os.path.join(input_type, f"{working_dir}-{unique_id}")

    doc_namespace = urlunparse(
        ("https", "fluidattacks.com", identifier, "", "", ""),
    )

    return doc_namespace


def package_to_spdx_pkg(package: Package) -> SPDX_Package:
    return SPDX_Package(
        name=package.name,
        download_location=NOASSERTION,
        spdx_id=get_spdx_id(package),
        version=package.version,
        license_concluded=NOASSERTION,
        primary_package_purpose=PackagePurpose.LIBRARY,
        external_references=[
            ExternalPackageRef(
                category=ExternalPackageRefCategory.PACKAGE_MANAGER,
                reference_type="purl",
                locator=package.p_url,
            )
        ],
    )


def format_spdx_sbom(  # pylint:disable=too-many-locals
    packages: list[Package],
    _relationships: list[Relationship],
    file_format: str,
    output: str,
    config: SbomConfig,
) -> None:
    now_utc = datetime.now(UTC)
    namespace, _ = set_namespace_version(config=config)
    creation_info = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name=namespace,
        data_license="CC0-1.0",
        document_namespace=document_namespace(namespace),
        creators=[Actor(ActorType.TOOL, "Fluid-Sbom", None)],
        created=now_utc,
    )
    document = Document(creation_info)

    package_cache = {pkg: package_to_spdx_pkg(pkg) for pkg in packages}

    spdx_packages: list[SPDX_Package] = list(package_cache.values())
    document.packages = spdx_packages

    doc_spdx_id = document.creation_info.spdx_id
    document_relationships = [
        SPDXRelationship(
            doc_spdx_id, SPDXRelationshipType.DESCRIBES, pkg.spdx_id
        )
        for pkg in spdx_packages
    ]

    for relationship in _relationships:
        to_pkg = package_cache.get(
            cast(Package, relationship.to_),
            package_to_spdx_pkg(cast(Package, relationship.to_)),
        )
        from_pkg = package_cache.get(
            cast(Package, relationship.from_),
            package_to_spdx_pkg(cast(Package, relationship.from_)),
        )

        document_relationships.append(
            SPDXRelationship(
                to_pkg.spdx_id,
                SPDXRelationshipType.DEPENDENCY_OF,
                from_pkg.spdx_id,
            )
        )

    document.relationships = document.relationships + document_relationships

    validation_errors: list[ValidationMessage] = validate_full_spdx_document(
        document
    )

    if validation_errors:
        raise SPDXValidationError(validation_errors)

    LOGGER.info(
        "🆗 SPDX %s valid, generating output file at %s.%s",
        file_format.upper(),
        output,
        file_format,
    )

    write_file(document, f"{output}.{file_format}")
    LOGGER.info("✅ Output file successfully generated")



from fluid_sbom.pkg.cataloger.java.archive_filename import (
    ArchiveFilename,
)
from fluid_sbom.pkg.java import (
    JavaManifest,
)
import unicodedata


def parse_line(
    line: str,
    last_key: str,
    current_section: int,
    sections: list[dict[str, str]],
) -> str:
    line = line.strip()

    # Handle continuation of the last key-value pair
    if line.startswith(" ") and last_key:
        sections[current_section][last_key] += " " + line.strip()
        return last_key

    # Handle new key-value pairs
    idx = line.find(":")
    if idx == -1:
        return last_key

    key = line[:idx].strip()
    value = line[idx + 1 :].strip()

    if not key:
        return last_key

    if last_key == "" or current_section == -1:
        sections.append({})
        current_section += 1

    sections[current_section][key] = value
    return key


def process_section(sections: list[dict[str, str]], reader: str) -> None:
    current_section = -1
    last_key = ""

    for line in reader.splitlines():
        if not line.strip():  # Empty lines denote section separators
            last_key = ""
            continue

        last_key = parse_line(line, last_key, current_section, sections)
        current_section = (
            len(sections) - 1
        )  # Update current section index after potential addition


def build_manifest(sections: list[dict[str, str]]) -> JavaManifest:
    if sections:
        main_section = sections[0]
        other_sections = sections[1:] if len(sections) > 1 else None
        return JavaManifest(main=main_section, sections=other_sections)
    return JavaManifest(main={}, sections=None)


def parse_java_manifest(reader: str) -> JavaManifest:
    sections: list[dict[str, str]] = []
    process_section(sections, reader)
    return build_manifest(sections)


def _field_value_from_manifest(manifest: JavaManifest, field: str) -> str:
    """
    Extract a field's value from a JavaManifest file.

    Args:
        manifest (JavaManifest): The JavaManifest file from where the field
            value has to be extracted.
        field (str): The specific field name which value is needed.

    Returns:
        str: The field value extracted from the Manifest file if it exists,
            else returns an empty string.
    """
    if (value := manifest.main.get(field, None)) and value:
        return value

    for section in manifest.sections or []:
        if (value := section.get(field, None)) and value:
            return value
    return ""


def select_licenses(manifest: JavaManifest) -> list[str]:
    """
    Select licenses from a JavaManifest

    Args:
        manifest (JavaManifest): the JavaManifest containing licenses

    Returns:
        list[str]: A list of licenses extracted from the manifest
    """
    result = []

    for field in (
        "Bundle-License",
        "Plugin-License-Name",
    ):
        if value := _field_value_from_manifest(manifest, field):
            result.append(value)

    return result


def extract_name_from_apache_maven_bundle_plugin(  # NOSONAR
    manifest: JavaManifest | None,
) -> str:
    """

    Extract the name from Apache Maven bundle plugin using manifest main
    information. If successful, the function tries to split the symbolic name
    according to convention `"${groupId}.${artifactId}"` and return the
    artifactId. If this process fails or the vendor id is equal to the
    symbolic name, an empty string is returned.

    Args:
        manifest (JavaManifest | None): The JavaManifest object containing the
        manifest main information or None.

    Returns:
        str: The extracted name from the Apache Maven bundle plugin, or an
        empty string if extraction fails.
    """
    if manifest and manifest.main:
        created_by = manifest.main.get("Created-By", "")
        if "Apache Maven Bundle Plugin" in created_by:
            symbolic_name = manifest.main.get("Bundle-SymbolicName", "")
            if symbolic_name:
                vendor_id = manifest.main.get("Implementation-Vendor-Id", "")
                if vendor_id and vendor_id == symbolic_name:
                    return ""

                # Assuming symbolicName convention "${groupId}.${artifactId}".
                fields = symbolic_name.split(".")
                # Potential issue with determining the actual artifactId based
                # on BND behavior.
                return fields[-1] if fields else ""

    return ""


def is_valid_java_identifier(field: str) -> bool:
    """
    Check if a string is a valid Java identifier.

    An identifier is a name assigned to elements in the Java program i.e.
    variable, method, class etc.

    Args:
        field (str): The string to be checked if it is a valid Java identifier.

    Returns:
        bool: True if the string is a valid Java identifier, False otherwise.
    """
    if not field:
        return False

    # Check the first character
    first_char = field[0]
    if first_char.isalpha() or unicodedata.category(first_char) in [
        "Sc",
        "Pc",
    ]:
        # If the first character is valid, check the remaining characters.
        # Note: Python's str.isidentifier() checks if the entire string is a
        # valid identifier.
        return field.isidentifier()

    return False


def extract_name_from_archive_filename(archive: ArchiveFilename) -> str:
    """
    Extracts name from an ArchiveFilename

    This function extracts the name from the provided ArchiveFilename.
    It handles special cases for `org.eclipse.*` group IDs. For valid Java
    identifiers, it returns the last part as the artifact ID.

    Args:
        archive (ArchiveFilename): Given archive filename from which the name
        needs to be extracted.

    Returns:
        str: The extracted name from the ArchiveFilename.
    """
    if "." in archive.name:
        # Handle the special case for 'org.eclipse.*' group IDs
        if archive.name.startswith("org.eclipse."):
            return archive.name

        # Split the name on dots and check if it looks like a
        # 'groupId.artifactId'
        fields = archive.name.split(".")
        if all(is_valid_java_identifier(f) for f in fields):
            # If all parts are valid Java identifiers, assume the last part
            # is the artifact ID
            return fields[-1]

    return archive.name


def select_name(
    manifest: JavaManifest | None, filename_obj: ArchiveFilename
) -> str:
    """
    Select a name from various sources based on the manifest and filename
    object

    Args:
        manifest (JavaManifest | None): The Java Manifest input from which to
        extract the name
        filename_obj (ArchiveFilename): The Archive Filename object

    Returns:
        str: Name extracted from the manifest or filename object. If no name
        is found in either of these sources, it tries to find a name from
        other attributes in the manifest. If no name is still found,
        returns an empty string.
    """
    name = extract_name_from_apache_maven_bundle_plugin(manifest)
    if name:
        return name

    # the filename tends to be the next-best reference for the package name
    name = extract_name_from_archive_filename(filename_obj)
    if name:
        return name

    # remaining fields in the manifest is a bit of a free-for-all depending on
    # the build tooling used and package maintainer preferences
    if manifest:
        main_attrs = manifest.main
        for key in [
            "Name",
            "Bundle-Name",
            "Short-Name",
            "Extension-Name",
            "Implementation-Title",
        ]:
            if key in main_attrs and main_attrs[key]:
                return main_attrs[key]

    return ""


def select_version(
    manifest: JavaManifest | None, filename_obj: ArchiveFilename
) -> str:
    """
    Selects the version from the Java manifest or the archive file

    Args:
        manifest (JavaManifest | None): The Java manifest file to extract
            version from, could be None
        filename_obj (ArchiveFilename): The archive file to extract version
            from

    Returns:
        str: The version extracted from either the manifest or the archive
            file, returns an empty string if no version can be found.
    """
    if version := filename_obj.version:
        return version

    if not manifest:
        return ""

    field_names = [
        "Implementation-Version",
        "Specification-Version",
        "Plugin-Version",
        "Bundle-Version",
    ]
    for field in field_names:
        if value := _field_value_from_manifest(manifest, field):
            return value

    return ""



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.linux.release import (
    Release,
)
from fluid_sbom.pkg.package import (
    Package,
)
from pydantic import (
    BaseModel,
    ConfigDict,
)
from typing import (
    Callable,
)


class Environment(BaseModel):
    linux_release: Release | None
    model_config = ConfigDict(frozen=True)


Parser = Callable[
    [Resolver, Environment, LocationReadCloser],
    tuple[list[Package], list[Relationship]] | None,
]



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
    Position,
)
from fluid_sbom.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.dart import (
    DartPubspecLickEntry,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)
from typing import (
    cast,
    ItemsView,
)
from urllib3.util import (
    parse_url,
)

LOGGER = logging.getLogger(__name__)


def get_hosted_url(entry: IndexedDict) -> str:
    hosted: str | None = entry.get("hosted")
    description: dict[str, str] | None = entry.get("description")

    if (
        hosted == "hosted"
        and description
        and description["url"] != "https://pub.dartlang.org"
    ):
        if host := parse_url(description["url"]).host:
            return host

        return description["url"]

    return ""


def get_vcs_url(entry: IndexedDict) -> str:
    source: str | None = entry.get("source")
    description: dict[str, str] | None = entry.get("description")

    if description and source == "git":
        if description.get("path") == ".":
            return f'{description["url"]}@{description["resolved-ref"]}'
        return (
            description["url"]
            + f'@{description["resolved-ref"]}'
            + f'#{description["path"]}'
        )
    return ""


def package_url(entry: DartPubspecLickEntry) -> str:
    qualifiers = {}
    if entry.hosted_url:
        qualifiers["hosted_url"] = entry.hosted_url
    elif entry.vcs_url:
        qualifiers["vcs_url"] = entry.vcs_url

    return PackageURL(  # type: ignore
        type="pub",
        namespace="",
        name=entry.name,
        version=entry.version,
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def parse_pubspec_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml: IndexedDict = cast(
        IndexedDict, parse_yaml_with_tree_sitter(reader.read_closer.read())
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []
    yaml_packages: IndexedDict = package_yaml["packages"]
    items: ItemsView[str, IndexedDict] = yaml_packages.items()
    for package_name, package_value in items:
        package_version: str = package_value["version"]
        current_location: Location = deepcopy(reader.location)
        if current_location.coordinates:
            position: Position = yaml_packages.get_key_position(package_name)
            current_location.coordinates.line = position.start.line
        metadata = DartPubspecLickEntry(
            name=package_name,
            version=package_version,
            hosted_url=get_hosted_url(package_value),
            vcs_url=get_vcs_url(package_value),
        )
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=package_version,
                    locations=[current_location],
                    language=Language.DART,
                    licenses=[],
                    type=PackageType.DartPubPkg,
                    p_url=package_url(metadata),
                    metadata=metadata,
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(  # type: ignore
                            include_url=False
                        ),
                        "location": current_location.path(),
                    }
                },
            )
            continue

    return packages, relationships



from datetime import (
    datetime,
)
from fluid_sbom.internal.package_information.java import (
    get_maven_package_info,
    MavenPackageInfo,
    search_maven_package,
)
from fluid_sbom.pkg.java import (
    JavaArchive,
    JavaPomProject,
    JavaPomProperties,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
from packageurl import (
    PackageURL,
)
from typing import (
    Any,
)


def looks_like_group_id(group_id: str) -> bool:
    return "." in group_id


def remove_osgi_directives(group_id: str) -> str:
    return group_id.split(";")[0]


def clean_group_id(group_id: str) -> str:
    return remove_osgi_directives(group_id).strip()


def group_id_from_pom_properties(
    properties: JavaPomProperties | None,
) -> str:
    if not properties:
        return ""
    if properties.group_id:
        return clean_group_id(properties.group_id)
    if properties.artifact_id and looks_like_group_id(properties.artifact_id):
        return clean_group_id(properties.artifact_id)

    return ""


def group_id_pom_project(project: JavaPomProject | None) -> str:
    if not project:
        return ""

    if project.group_id:
        return clean_group_id(project.group_id)

    if project.artifact_id and looks_like_group_id(project.artifact_id):
        return clean_group_id(project.artifact_id)

    if project.parent:
        if project.parent.group_id:
            return clean_group_id(project.parent.group_id)

        if looks_like_group_id(project.parent.artifact_id):
            return clean_group_id(project.parent.artifact_id)

    return ""


def group_id_from_java_metadata(_pkg_name: str, metadata: Any) -> str | None:
    if hasattr(metadata, "pom_properties") and (
        group_id := group_id_from_pom_properties(metadata.pom_properties)
    ):
        return group_id

    if hasattr(metadata, "pom_project") and (
        group_id := group_id_pom_project(metadata.pom_project)
    ):
        return group_id

    return None


def package_url(name: str, version: str, metadata: JavaArchive) -> str:
    group_id = name
    if (g_id := group_id_from_java_metadata(name, metadata)) and g_id:
        group_id = g_id
    return PackageURL(
        type="maven",
        namespace=group_id,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _set_health_metadata_artifact(
    package: Package,
    current_package: MavenPackageInfo | None,
) -> Package:
    if current_package and package.health_metadata and current_package.jar_url:
        digest_value = current_package.hash or None
        package.health_metadata.artifact = Artifact(
            url=current_package.jar_url,
            integrity=Digest(
                algorithm="sha1" if digest_value else None,
                value=digest_value,
            ),
        )
    return package


def _set_licenses(
    package: Package, maven_package: MavenPackageInfo
) -> Package:
    if not package.licenses:
        package.licenses = maven_package.licenses or []
    return package


def _set_authors(package: Package, maven_package: MavenPackageInfo) -> Package:
    if package.health_metadata and not package.health_metadata.authors:
        authors = maven_package.authors or []
        package.health_metadata.authors = (
            ", ".join(authors) if authors else None
        )
    return package


def complete_package(package: Package) -> Package:
    current_package: MavenPackageInfo | None = None
    group_id: str | None = None
    artifact_id: str = package.name
    version: str = package.version

    if g_id := group_id_from_java_metadata(package.name, package.metadata):
        group_id = g_id
    elif package_candidate := search_maven_package(
        package.name, package.version
    ):
        group_id = package_candidate.group

    if not group_id:
        return package

    maven_package = get_maven_package_info(group_id, package.name)
    if not maven_package and (
        package_candidate := search_maven_package(
            package.name, package.version
        )
    ):
        maven_package = get_maven_package_info(
            package_candidate.group, package.name
        )

    if not maven_package:
        return package
    if maven_package.latest_version and maven_package.release_date:
        try:
            package.health_metadata = HealthMetadata(
                latest_version=maven_package.latest_version,
                latest_version_created_at=datetime.fromtimestamp(
                    maven_package.release_date
                ),
            )
        except IndexError:
            return package
    if version and (
        current_package := get_maven_package_info(
            group_id,
            artifact_id,
            version,
        )
    ):
        package = _set_health_metadata_artifact(package, current_package)

    package = _set_authors(package, maven_package)
    package = _set_licenses(package, maven_package)

    return package



import bugsnag
from bugsnag_client import (
    add_batch_metadata as bugsnag_add_batch_metadata,
    remove_nix_hash as bugsnag_remove_nix_hash,
)
from fluid_sbom.context import (
    BASE_DIR,
    CI_COMMIT_SHORT_SHA,
)
from fluid_sbom.utils.env import (
    guess_environment,
)
import logging


def initialize_bugsnag() -> None:
    bugsnag.before_notify(bugsnag_add_batch_metadata)
    bugsnag.before_notify(bugsnag_remove_nix_hash)
    bugsnag.configure(
        ignore_classes=["SystemExit"],
        breadcrumb_log_level=logging.DEBUG,
        notify_release_stages=["production"],
        release_stage=guess_environment(),
        app_version=CI_COMMIT_SHORT_SHA,
        project_root=BASE_DIR,
    )
    bugsnag.start_session()



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal import (
    licenses,
)
from fluid_sbom.internal.digest import (
    new_digests_from_file,
)
from fluid_sbom.internal.file.zip_file_manifest import (
    new_zip_file_manifest,
    zip_glob_match,
)
from fluid_sbom.internal.file.zip_file_traversal import (
    contents_from_zip,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.java.archive_filename import (
    ArchiveFilename,
    parse_filename,
)
from fluid_sbom.pkg.cataloger.java.package import (
    group_id_from_java_metadata,
    package_url,
)
from fluid_sbom.pkg.cataloger.java.parse_java_manifest import (
    parse_java_manifest,
    select_licenses,
    select_name,
    select_version,
)
from fluid_sbom.pkg.cataloger.java.parse_pom_properties import (
    parse_pom_properties,
)
from fluid_sbom.pkg.cataloger.java.parse_pom_xml import (
    parse_pom_xml_project,
    ParsedPomProject,
)
from fluid_sbom.pkg.java import (
    JavaArchive,
    JavaManifest,
    JavaPomProject,
    JavaPomProperties,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
import logging
import os
from pathlib import (
    Path,
)
from pydantic import (
    BaseModel,
    ConfigDict,
    ValidationError,
)
import shutil
import tempfile
from typing import (
    Callable,
    cast,
    TextIO,
)
from zipfile import (
    ZipInfo,
)

LOGGER = logging.getLogger(__name__)


def get_digests_from_archive(archive_path: str) -> list[Digest]:
    with open(archive_path, "rb") as reader:
        return new_digests_from_file(reader, ["sha1"])


def pom_properties_by_parent(
    archive_path: str, location: Location, extract_paths: list[str]
) -> dict[str, JavaPomProperties]:
    properties_by_parent_path = {}
    contents_of_maven_properties = contents_from_zip(
        archive_path, *extract_paths
    )

    for file_path, file_contents in contents_of_maven_properties.items():
        if not file_contents:
            logging.error(
                "failed to parse pom.properties",
                extra={
                    "contents_path": file_path,
                    "location": location.coordinates.real_path
                    if location.coordinates
                    else location.access_path,
                },
            )
            continue
        pom_properties = parse_pom_properties(file_path, file_contents)
        if not pom_properties:
            continue
        if not pom_properties.group_id or not pom_properties.version:
            continue
        properties_by_parent_path[str(Path(file_path).parent)] = pom_properties

    return properties_by_parent_path


def pom_project_by_parent(
    archive_path: str, location: Location, extract_paths: list[str]
) -> dict[str, ParsedPomProject]:
    contents_of_maven_project = contents_from_zip(archive_path, *extract_paths)

    project_by_parent = {}

    for file_path, file_contents in contents_of_maven_project.items():
        pom_project = parse_pom_xml_project(file_path, file_contents, location)
        if not pom_project:
            continue
        if (
            not pom_project.java_pom_project.parent
            and not pom_project.java_pom_project.version
        ) or not pom_project.java_pom_project.artifact_id:
            continue

        project_by_parent[str(Path(file_path).parent)] = pom_project

    return project_by_parent


def artifact_id_matches_filename(artifact_id: str, filename: str) -> bool:
    if not artifact_id or not filename:
        return False

    return artifact_id.startswith(filename) or filename.startswith(artifact_id)


class ArchiveParser(BaseModel):
    file_manifest: list[ZipInfo]
    location: Location
    archive_path: str | None
    content_path: str
    file_info: ArchiveFilename
    detect_nested: bool
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parse(self) -> tuple[list[Package] | None, list[Relationship] | None]:
        packages = []
        parent_pkg = self.discover_main_package()
        if not parent_pkg:
            return None, None
        aux_pkgs = self.discover_pkgs_from_all_maven_files(parent_pkg)

        packages = [parent_pkg, *aux_pkgs]

        return packages, []

    def discover_pkgs_from_all_maven_files(
        self, parent_pkg: Package | None
    ) -> list[Package]:
        if not parent_pkg or not self.archive_path:
            return []

        pkgs: list[Package] = []

        properties = pom_properties_by_parent(
            self.archive_path,
            self.location,
            zip_glob_match(self.file_manifest, False, "*pom.properties"),
        )

        projects = pom_project_by_parent(
            self.archive_path,
            self.location,
            zip_glob_match(self.file_manifest, False, "*pom.xml"),
        )

        for parent_path, properties_obj in properties.items():
            pom_project: ParsedPomProject | None = projects.get(
                parent_path, None
            )
            if pkg_from_pom := new_package_from_maven_data(
                properties_obj,
                pom_project,
                parent_pkg,
                self.location,
            ):
                pkgs.append(pkg_from_pom)
        return pkgs

    def discover_main_package(self) -> Package | None:
        manifest_matches = zip_glob_match(
            self.file_manifest, False, "/META-INF/MANIFEST.MF"
        )
        if not manifest_matches or not self.archive_path:
            logging.error("No manifest found in the jar")
            return None
        contents = contents_from_zip(self.archive_path, *manifest_matches)
        if not contents:
            logging.error("Unable to read manifest contents")
            return None

        manifest_contents = contents[manifest_matches[0]]
        manifest = parse_java_manifest(manifest_contents)
        licenses_, name, version = self.parse_licenses(manifest)
        metadata = JavaArchive(
            virtual_path=self.location.path(),
            manifest=manifest,
            archive_digests=get_digests_from_archive(self.archive_path),
        )
        try:
            return Package(
                name=name,
                version=version,
                licenses=licenses_,
                locations=[self.location],
                type=self.file_info.pkg_type(),
                language=Language.JAVA,
                metadata=metadata,
                p_url=package_url(name, version, metadata),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data "
                "types are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": self.location.path(),
                    }
                },
            )
            return None

    def get_license_from_file_in_archive(self) -> list[str]:
        file_licenses = []
        for (
            filename
        ) in (
            licenses.LICENSES_FILE_NAMES
        ):  # Assuming this function returns a list of filenames to check
            license_matches = zip_glob_match(
                self.file_manifest, True, f"/META-INF/{filename}"
            )
            if not license_matches:
                # Try the root directory if it's not in META-INF
                license_matches = zip_glob_match(
                    self.file_manifest, True, f"/{filename}"
                )

            if license_matches and self.archive_path:
                contents = contents_from_zip(
                    self.archive_path, *license_matches
                )
                for license_match in license_matches:
                    license_contents = contents.get(license_match, "")

                    parsed = licenses.parse_license(
                        license_contents
                    )  # Assuming this function parses license contents
                    if parsed:
                        file_licenses.extend(parsed)

        return file_licenses

    def parse_licenses(
        self, manifest: JavaManifest
    ) -> tuple[list[str], str, str]:
        licenses_ = select_licenses(manifest)

        (
            name,
            version,
            pom_licenses,
        ) = self.guest_main_package_name_and_version_from_po29m()

        if not name:
            name = select_name(manifest, self.file_info)
        if not version:
            version = select_version(manifest, self.file_info)

        if not licenses_:
            licenses_.extend(pom_licenses or [])

        if not licenses_:
            file_licenses = self.get_license_from_file_in_archive()
            if file_licenses:
                licenses_.extend(file_licenses)
        return licenses_, name, version

    def extract_properties_and_projects(
        self,
    ) -> tuple[dict[str, JavaPomProperties], dict[str, ParsedPomProject]]:
        properties = {}
        projects = {}

        pom_property_matches = zip_glob_match(
            self.file_manifest, False, "*pom.properties"
        )
        pom_matches = zip_glob_match(self.file_manifest, False, "*pom.xml")
        if self.archive_path:
            properties = pom_properties_by_parent(
                self.archive_path, self.location, pom_property_matches
            )
            projects = pom_project_by_parent(
                self.archive_path, self.location, pom_matches
            )

        return properties, projects

    def find_relevant_objects(
        self,
        properties: dict[str, JavaPomProperties],
        projects: dict[str, ParsedPomProject],
    ) -> tuple[JavaPomProperties | None, ParsedPomProject | None]:
        for parent_path, properties_obj in properties.items():
            if properties_obj.artifact_id and artifact_id_matches_filename(
                properties_obj.artifact_id, self.file_info.name
            ):
                if proj := projects.get(parent_path):
                    return properties_obj, proj
        return None, None

    def extract_name_version(
        self, properties_obj: JavaPomProperties, project_obj: ParsedPomProject
    ) -> tuple[str | None, str | None]:
        name = properties_obj.artifact_id if properties_obj else None
        version = properties_obj.version if properties_obj else None

        if not name and project_obj:
            name = project_obj.java_pom_project.artifact_id
        if not version and project_obj:
            version = project_obj.java_pom_project.version

        return name, version

    def guest_main_package_name_and_version_from_po29m(
        self,
    ) -> tuple[str | None, str | None, list[str] | None]:
        properties, projects = self.extract_properties_and_projects()
        properties_obj, project_obj = self.find_relevant_objects(
            properties, projects
        )

        if not properties_obj or not project_obj:
            return None, None, None

        name, version = self.extract_name_version(properties_obj, project_obj)

        return name, version, []


def save_archive_to_tmp(
    archive_virtual_path: str, _reader: TextIO
) -> tuple[str | None, str | None, Callable[[], None]]:
    name = os.path.basename(archive_virtual_path)
    temp_dir = tempfile.mkdtemp(prefix="sbom-archive-contents-")

    def cleanup_fn() -> None:
        shutil.rmtree(temp_dir)

    content_dir = os.path.join(temp_dir, "contents")

    os.mkdir(content_dir)

    archive_path = os.path.join(temp_dir, f"archive-{name}")
    shutil.copy(archive_virtual_path, archive_path)

    return content_dir, archive_path, cleanup_fn


def new_java_archive_parser(
    reader: LocationReadCloser, detect_nested: bool
) -> tuple[ArchiveParser | None, Callable[[], None] | None]:
    if not reader.location.coordinates:
        return None, None
    current_file_path = reader.location.coordinates.real_path
    content_path, archive_path, cleanup_fn = save_archive_to_tmp(
        current_file_path, reader.read_closer
    )
    if not archive_path or not content_path:
        logging.error("unable to read files from java archive")
        return None, None

    file_info = parse_filename(current_file_path)
    file_manifest = new_zip_file_manifest(archive_path)
    return (
        ArchiveParser(
            file_manifest=file_manifest,
            location=reader.location,
            archive_path=archive_path,
            content_path=content_path,
            file_info=file_info,
            detect_nested=detect_nested,
        ),
        cleanup_fn,
    )


def parse_java_archive(
    _: Resolver, __: Environment, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    parser, _file_cleanup_fn = new_java_archive_parser(
        reader, detect_nested=True
    )
    if parser is None:
        return [], []

    pkgs, relations = parser.parse()
    return pkgs or [], relations or []


def new_package_from_maven_data(
    pom_properties: JavaPomProperties,
    _parsed_pom_project: ParsedPomProject | None,
    parent_pkg: Package,
    location: Location,
) -> Package | None:
    v_path_suffix = ""

    parent_metadata: JavaArchive = cast(JavaArchive, parent_pkg.metadata)
    group_id = group_id_from_java_metadata(parent_pkg.name, parent_metadata)
    parent_key = f"{group_id}:{parent_pkg.name}:{parent_pkg.version}"
    pom_project_key = (
        f"{pom_properties.group_id}:"
        f"{pom_properties.artifact_id}:{pom_properties.version}"
    )

    if parent_key != pom_project_key:
        v_path_suffix += (
            f":{pom_properties.group_id}:{pom_properties.artifact_id}"
        )

    virtual_path = f"{location.path()}{v_path_suffix}"
    pkg_pom_project: JavaPomProject | None = None

    metadata = JavaArchive(
        virtual_path=virtual_path,
        pom_properties=pom_properties,
        pom_project=pkg_pom_project,
        parent=parent_pkg,
    )

    if not pom_properties.artifact_id or not pom_properties.version:
        return None

    try:
        return Package(
            name=pom_properties.artifact_id,
            version=pom_properties.version,
            licenses=[],
            locations=[location],
            type=pom_properties.pkg_type_indicated(),
            language=Language.JAVA,
            metadata=metadata,
            p_url=package_url(
                pom_properties.artifact_id, pom_properties.version, metadata
            ),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data "
            "types are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": location.path(),
                }
            },
        )
        return None



from fluid_sbom.file.resolver import (
    Resolver,
)
import os_release
from os_release.parser import (
    OsReleaseParseException,
)
from pydantic import (
    BaseModel,
)


class Release(BaseModel):
    id_: str
    version_id: str
    name: str | None = None
    pretty_name: str | None = None
    version: str | None = None
    id_like: list[str] | None = None
    version_code_name: str | None = None
    build_id: str | None = None
    image_id: str | None = None
    image_version: str | None = None
    variant: str | None = None
    variant_id: str | None = None
    home_url: str | None = None
    support_url: str | None = None
    bug_report_url: str | None = None
    privacy_policy_url: str | None = None
    cpe_name: str | None = None
    support_end: str | None = None

    def __str__(self) -> str:
        if self.pretty_name:
            return self.pretty_name
        if self.name:
            return self.name
        if self.version:
            return f"{self.id_} {self.version}"
        if self.version_id != "":
            return f"{self.id_} {self.version_id}"
        return f"{self.id_} {self.build_id or ''}"


def _force_parse(content: str) -> dict[str, str] | None:
    lines: list[tuple[str, str]] = []
    for line in content.splitlines():
        line = line.rstrip("\n")
        line = line.strip()
        if not line:
            continue
        pair = line.split("=", 1)
        if len(pair) != 2:
            continue
        pair[-1] = pair[-1].strip("\"'")
        lines.append((pair[0], pair[1]))
    if not lines:
        return None

    return dict(lines)


def parse_os_release(content: str) -> Release | None:
    try:
        release = os_release.parse_str(content)
    except OsReleaseParseException:
        release = _force_parse(content)
    if release:
        if "ID_LIKE" in release:
            release["ID_LIKE"] = sorted(release["ID_LIKE"].split(" "))
        return Release(
            pretty_name=release.get("PRETTY_NAME", ""),
            name=release.get("NAME", ""),
            id_=release.get("ID", ""),
            id_like=release.get("ID_LIKE", []),
            version=release.get("VERSION", ""),
            version_id=release.get("VERSION_ID", ""),
            version_code_name=release.get("VERSION_CODENAME", ""),
            build_id=release.get("BUILD_ID", ""),
            image_id=release.get("IMAGE_ID", ""),
            image_version=release.get("IMAGE_VERSION", ""),
            variant=release.get("VARIANT", ""),
            variant_id=release.get("VARIANT_ID", ""),
            home_url=release.get("HOME_URL", ""),
            support_url=release.get("SUPPORT_URL", ""),
            bug_report_url=release.get("BUG_REPORT_URL", ""),
            privacy_policy_url=release.get("PRIVACY_POLICY_URL", ""),
            cpe_name=release.get("CPE_NAME", ""),
            support_end=release.get("SUPPORT_END", ""),
        )
    return None


def identify_release(resolver: Resolver) -> Release | None:
    possible_files = [
        "/etc/os-release",
        "/usr/lib/os-release",
        "/etc/system-release-cpe",
        "/etc/redhat-release",
        "/bin/busybox",
    ]

    for file in possible_files:
        if not resolver.has_path(file):
            continue
        location = resolver.files_by_path(file)[0]
        content_reader = resolver.file_contents_by_location(location)
        if not content_reader:
            continue
        content = content_reader.read()
        release = parse_os_release(content)
        if release:
            return release
    return None



from fluid_sbom.sources.docker import (
    extract_docker_image,
    ImageMetadata,
)
import os
from pydantic import (
    BaseModel,
)
import re
import tempfile
from typing import (
    Any,
)


class LayerInfo(BaseModel):
    id_: str
    tar_full_path: str
    metadata: dict[str, Any]


class ImageContext(BaseModel):
    id: str
    name: str
    publisher: str | None
    arch: str
    size: str
    full_extraction_dir: str
    layers_dir: str
    manifest: dict[str, Any]
    image_ref: str

    def get_layer_info(self, layer_id: str) -> LayerInfo | None:
        if layer_metadata := next(
            x for x in self.manifest["Layers"] if x["id"] == layer_id
        ):
            return LayerInfo(
                id_=layer_id,
                tar_full_path=os.path.join(self.layers_dir, layer_metadata),
                metadata=layer_metadata,
            )
        return None


def get_context(
    image: ImageMetadata,
    *,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    daemon: bool = False,
) -> ImageContext | None:
    temp_dir = tempfile.mkdtemp()
    layers_dir, manifest = extract_docker_image(
        image,
        temp_dir,
        username=username,
        password=password,
        token=token,
        daemon=daemon,
    )

    return ImageContext(
        id=image.digest,
        name=image.name,
        publisher="",
        arch=image.architecture,
        size=str(sum(x.size for x in image.layersdata)),
        full_extraction_dir=temp_dir,
        layers_dir=layers_dir,
        manifest=manifest,
        image_ref=image.image_ref,
    )


def extract_image_ref_info(user_input: str) -> dict[str, str] | None:
    patter = (
        r"^((?P<publisher>[a-z0-9]+(?:[._-][a-z0-9]+)*)\/)?"
        r"(?P<name>[a-z0-9]+(?:[._-][a-z0-9]+)*)"
        r"(:(?P<tag>[a-zA-Z0-9_.-]+))?"
        r"(?P<digest>@sha256:[a-fA-F0-9]{64})?$"
    )
    if match := re.match(patter, user_input):
        return match.groupdict()
    return None



from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.linux.release import (
    Release,
)
from fluid_sbom.pkg.alpm import (
    AlpmDBEntry,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.pkg.url import (
    purl_qualifiers,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def package_url(entry: AlpmDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": entry.architecture}
    if entry.base_package:
        qualifiers["upstream"] = entry.base_package
    return PackageURL(
        type="alpm",
        name=entry.package,
        version=entry.version,
        qualifiers=purl_qualifiers(qualifiers, distro),  # type: ignore
        subpath="",
    ).to_string()


def new_package(
    entry: AlpmDBEntry, release: Release | None, db_location: Location
) -> Package | None:
    licenses_candidates = entry.licenses.split("\n")
    try:
        return Package(
            name=entry.package,
            version=entry.version,
            locations=[db_location],
            licenses=licenses_candidates,
            type=PackageType.AlpmPkg,
            metadata=entry,
            p_url=package_url(entry, release),
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data "
            "types are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": db_location.path(),
                }
            },
        )
        return None



from fluid_sbom.pkg.package import (
    Advisory,
    Package,
)
import logging
from packageurl import (
    PackageURL,
)
import re
from typing import (
    Self,
)
from univers.version_range import (
    InvalidVersionRange,
    RANGE_CLASS_BY_SCHEMES,
    VersionConstraint,
    VersionRange,
)
from univers.versions import (
    AlpineLinuxVersion,
    ComposerVersion,
    DebianVersion,
    GolangVersion,
    InvalidVersion,
    MavenVersion,
    NugetVersion,
    PypiVersion,
    RpmVersion,
    RubygemsVersion,
    SemverVersion,
    Version,
)

LOGGER = logging.getLogger(__name__)


def _get_version_scheme_by_namespace(
    package: Package, namespace: str
) -> Version | None:
    schemes = {
        "distro": {
            "alpine": AlpineLinuxVersion,
            "debian": DebianVersion,
            "redhat": RpmVersion,
            "ubuntu": DebianVersion,
        },
        "language": {
            "dart": SemverVersion,
            "dotnet": NugetVersion,
            "go": GolangVersion,
            "java": MavenVersion,
            "javascript": SemverVersion,
            "php": ComposerVersion,
            "python": PypiVersion,
            "ruby": RubygemsVersion,
            "rust": SemverVersion,
            "swift": SemverVersion,
        },
    }

    def _get_language_scheme() -> Version | None:
        return schemes["language"].get(package.language.value)

    def _get_distro_scheme() -> Version | None:
        if package.p_url:
            package_url = PackageURL.from_string(package.p_url)
            if isinstance(package_url.qualifiers, dict) and (
                distro := package_url.qualifiers.get("distro_id")
            ):
                return schemes["distro"].get(distro)
        return None

    parts = namespace.split(":")
    if len(parts) < 3:
        return _get_language_scheme() or _get_distro_scheme()

    namespace_type, subtype = parts[1], parts[2]
    result = schemes.get(namespace_type, {}).get(subtype)

    return result or _get_language_scheme() or _get_distro_scheme()


class ApkVersionRange(VersionRange):  # type: ignore[misc]
    scheme = "apk"
    version_class = AlpineLinuxVersion

    @classmethod
    def from_native(cls, string: str) -> Self:
        match = re.match(r"([<>=~!^]*)(.*)", string)
        if not match:
            raise ValueError(f"Invalid version range: {string}")
        comparator, version = match.groups()
        version = version.strip()
        return cls(
            constraints=[
                VersionConstraint(
                    comparator=comparator, version=cls.version_class(version)
                ),
            ]
        )


def _compare_single_constraint(
    version: Version,
    constraint: str,
    scheme: str,
) -> bool:
    version_range: VersionRange | None = {
        **RANGE_CLASS_BY_SCHEMES,
        "apk": ApkVersionRange,
    }.get(scheme)

    if not version_range:
        raise ValueError(f"Invalid version scheme: {scheme}")
    try:
        return version in version_range.from_native(constraint.strip())
    except (InvalidVersion, InvalidVersionRange, TypeError):
        return False


def _matches_constraint(
    version: Version, constraint: str, version_scheme: str
) -> bool:
    if not constraint:
        return True

    constraints = constraint.split("||")
    return any(
        _compare_single_constraint(version, constraint.strip(), version_scheme)
        for constraint in constraints
    )


def matches_version(package: Package, advisory: Advisory) -> bool:
    version_type = _get_version_scheme_by_namespace(
        package, advisory.namespace
    )
    if version_type is None:
        LOGGER.debug(
            "No version scheme found for namespace %s",
            advisory.namespace,
        )
        return False

    if advisory.version_constraint is None:
        return True
    if not package.p_url:
        return False

    try:
        match = re.match(r"([<>=~!^]*)(.*)", package.version)
        if not match:
            return False

        _, version = match.groups()
        version = version.strip()

        return all(
            _matches_constraint(
                version_type(version),
                constraint,
                PackageURL.from_string(package.p_url).type,
            )
            for constraint in advisory.version_constraint.split(",")
        )
    except (AttributeError, InvalidVersion):
        return False



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.toml import (
    parse_toml_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.rust.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.rust import (
    RustCargoLockEntry,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def _create_package(pkg: IndexedDict, location: Location) -> Package | None:
    name: str = pkg["name"]
    version: str = pkg["version"]
    source: str | None = pkg.get("source")
    dependencies: list[str] | None = pkg.get("dependencies")
    checksum: str | None = pkg.get("checksum")

    current_location = deepcopy(location)
    if current_location.coordinates and hasattr(pkg, "position"):
        current_location.coordinates.line = pkg.position.start.line

    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.RUST,
            licenses=[],
            p_url=package_url(name, version),
            type=PackageType.RustPkg,
            metadata=RustCargoLockEntry(
                name=name,
                version=version,
                source=source,
                dependencies=dependencies or [],
                checksum=checksum,
            ),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),  # type: ignore
                    "location": current_location.path(),
                }
            },
        )
        return None


def _create_relationships(packages: list[Package]) -> list[Relationship]:
    relationships = []
    for pkg in packages:
        if isinstance(pkg.metadata, RustCargoLockEntry):
            for dep_name in pkg.metadata.dependencies:
                if dep := next(
                    (x for x in packages if x.name == dep_name), None
                ):
                    relationships.append(
                        Relationship(
                            from_=pkg,
                            to_=dep,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                            data=None,
                        )
                    )
    return relationships


def parse_cargo_lock(
    _: Resolver | None, __: Environment | None, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    if not reader.location.coordinates:
        return packages, []

    _content = reader.read_closer.read()
    try:
        toml = parse_toml_with_tree_sitter(_content)
    except UnexpectedNode:
        return [], []

    toml_pkgs: list[IndexedDict] | None = toml.get("package")
    for pkg in toml_pkgs or []:
        if package := _create_package(pkg, reader.location):
            packages.append(package)

    relationships = _create_relationships(packages)

    return packages, relationships



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.php.package import (
    new_package_from_composer,
)
from fluid_sbom.pkg.package import (
    Package,
)
import logging
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def _load_package_json(
    reader: LocationReadCloser,
) -> IndexedDict | None:
    try:
        return cast(
            IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
        )
    except ValueError:
        LOGGER.warning(
            "Failed to parse package.json",
            extra={"extra": {"location": reader.location}},
        )

        return None


def _extract_packages(
    package_json: IndexedDict, location: Location
) -> list[Package]:
    packages = []
    dev_packages = (
        package_json["dev-package-names"]
        if isinstance(package_json, IndexedDict)
        else []
    )
    for package in (
        package_json["packages"]
        if "packages" in package_json
        else package_json
    ):
        pkg_item = new_package_from_composer(
            package, deepcopy(location), False
        )
        if pkg_item:
            pkg_item.is_dev = pkg_item.name in dev_packages
            packages.append(pkg_item)

    return packages


def _extract_relationships(packages: list[Package]) -> list[Relationship]:
    relationships = []
    for package in packages:
        package_metadata = getattr(package.metadata, "require", None)
        dependencies = (
            list(package_metadata.keys()) if package_metadata else []
        )
        for dep_name in dependencies:
            package_dep = next(
                (x for x in packages if x.name == dep_name), None
            )
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=package,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data=None,
                    )
                )

    return relationships


def parse_installed_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = _load_package_json(reader)

    if not package_json:
        return [], []

    packages = _extract_packages(package_json, reader.location)
    relationships = _extract_relationships(packages)

    return packages, relationships



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.java.package import (
    package_url,
)
from fluid_sbom.pkg.java import (
    JavaArchive,
    JavaPomProject,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from pydantic import (
    BaseModel,
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def parse_gradle_lockfile(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies: list[LockFileDependency] = []
    packages: list[Package] = []
    for line_number, line in enumerate(reader.read_closer.readlines(), 1):
        if "=" in line and ":" in line:  # To ensure it's a dependency line
            dependency_part = line.split("=")[0]
            group, name, version = dependency_part.split(":")
            dependencies.append(
                LockFileDependency(
                    group=group, name=name, version=version, line=line_number
                )
            )

    for dependency in dependencies:
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = JavaArchive(
            pom_project=JavaPomProject(
                group_id=dependency.group,
                name=dependency.name,
                artifact_id=dependency.name,
                version=dependency.version,
            )
        )
        try:
            packages.append(
                Package(
                    name=dependency.name,
                    version=dependency.version,
                    locations=[location],
                    language=Language.JAVA,
                    type=PackageType.JavaPkg,
                    metadata=archive,
                    p_url=package_url(
                        dependency.name, dependency.version, archive
                    ),
                    licenses=[],
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue

    return packages, []



from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.pkg.elixir import (
    ElixirMixLockEntry,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def new_package(
    entry: ElixirMixLockEntry, locations: Location
) -> Package | None:
    try:
        return Package(
            name=entry.name,
            version=entry.version,
            type=PackageType.HexPkg,
            locations=[locations],
            p_url=package_url(entry),
            metadata=entry,
            language=Language.ELIXIR,
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": locations.path(),
                }
            },
        )
        return None


def package_url(entry: ElixirMixLockEntry) -> str:
    return PackageURL(
        "hex", "", entry.name, entry.version, None, ""
    ).to_string()



from datetime import (
    datetime,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.file import (
    Digest,
)
import hashlib
import json
from pydantic import (
    BaseModel,
    Field,
    field_validator,
)


class Advisory(BaseModel):
    cpes: list[str]
    description: str | None = Field(min_length=1)
    epss: float
    id: str = Field(min_length=1)
    namespace: str = Field(min_length=1)
    percentile: float
    severity: str = Field(min_length=1)
    urls: list[str]
    version_constraint: str | None = Field()

    @field_validator("cpes")
    @classmethod
    def check_cpes_min_length(cls, value: list[str]) -> list[str]:
        for cpe in value:
            if len(cpe) < 1:
                raise ValueError(
                    "Each cpe string must be at least 1 character long."
                )
        return value

    @field_validator("urls")
    @classmethod
    def check_urls_min_length(cls, value: list[str]) -> list[str]:
        for url in value:
            if len(url) < 1:
                raise ValueError(
                    "Each url string must be at least 1 character long."
                )
        return value

    def get_info_count(self) -> int:
        info_count = 0
        for attr in [
            self.cpes,
            self.description,
            self.urls,
            self.version_constraint,
        ]:
            if attr:
                info_count += 1
        return info_count

    def __repr__(self) -> str:
        return (
            f"Advisory(id={self.id}, "
            f"namespace={self.namespace}, severity={self.severity})"
        )


class Artifact(BaseModel):
    url: str = Field(min_length=1)
    integrity: Digest | None = None


class HealthMetadata(BaseModel):
    latest_version: str | None = Field(default=None, min_length=1)
    latest_version_created_at: str | datetime | None = None
    artifact: Artifact | None = None
    authors: str | None = Field(default=None, min_length=1)

    @field_validator("latest_version_created_at", mode="before")
    @classmethod
    def validate_latest_version_created_at(
        cls, value: str | datetime | None
    ) -> str | datetime | None:
        if isinstance(value, str) and len(value) < 1:
            raise ValueError(
                "latest_version_created_at must be at least 1 character long "
                "when it is a string."
            )
        return value


class Package(BaseModel):
    name: str = Field(min_length=1)
    version: str = Field(min_length=1)
    language: Language
    licenses: list[str]
    locations: list[Location]
    type: PackageType
    advisories: list[Advisory] | None = None
    dependencies: list["Package"] | None = None
    found_by: str | None = Field(default=None, min_length=1)
    health_metadata: HealthMetadata | None = None
    is_dev: bool = False
    metadata: object | None = None
    p_url: str = Field(min_length=1)

    @property
    def id_(self) -> str:
        return self.id_by_hash()

    def id_by_hash(self) -> str:
        try:
            obj_data = {
                "name": self.name,
                "version": self.version,
                "language": self.language.value,
                "type": self.type.value,
                "p_url": self.p_url,
            }
            obj_str = json.dumps(obj_data, sort_keys=True)
            hash_value = hashlib.sha256(obj_str.encode()).hexdigest()
            return hash_value
        except Exception as exc:  # pylint:disable=broad-exception-caught
            return f"could not build ID for object={self}: {exc}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Package):
            return self.id_ == other.id_
        return False

    def __hash__(self) -> int:
        return hash(self.id_)

    @field_validator("licenses")
    @classmethod
    def check_licenses_min_length(cls, value: list[str]) -> list[str]:
        for license_str in value:
            if len(license_str) < 1:
                raise ValueError(
                    "Each license string must be at least 1 character long."
                )
        return value



from fluid_sbom.file.location import (
    Location,
)
from io import (
    TextIOWrapper,
)
from pydantic import (
    BaseModel,
    ConfigDict,
)
from typing import (
    TextIO,
)


class LocationReadCloser(BaseModel):
    location: Location
    read_closer: TextIO | TextIOWrapper
    model_config = ConfigDict(arbitrary_types_allowed=True)



from fluid_sbom.internal.cache import (
    dual_cache,
)
import requests
from typing import (
    Any,
)


@dual_cache
def get_composer_package(
    package_name: str, version: str | None = None
) -> dict[str, Any] | None:
    base_url = f"https://repo.packagist.org/p2/{package_name}.json"
    response = requests.get(base_url, timeout=20)
    if response.status_code == 200:
        package_data = response.json()
        if version:
            # Search for the specific version in the package data
            for package_version in package_data["packages"][package_name]:
                if package_version["version"] == version:
                    return package_version
            return None  # Version not found

        latest_version = package_data["packages"][package_name][0]
        return latest_version

    return None



from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection import (
    toml,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.python import (
    PythonRequirementsEntry,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def parse_poetry_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    _content = reader.read_closer.read()
    try:
        toml_content: IndexedDict = toml.parse_toml_with_tree_sitter(_content)
    except UnexpectedNode:
        return [], []
    for package in toml_content.get("package", []):
        p_url = PackageURL(
            type="pypi",
            namespace="",
            name=package["name"],
            version=package["version"],
            qualifiers="",
            subpath="",
        ).to_string()
        try:
            packages.append(
                Package(
                    name=package["name"],
                    version=package["version"],
                    found_by=None,
                    locations=[
                        (
                            _get_location(
                                reader.location, package.position.start.line
                            )
                            if isinstance(package, IndexedDict)
                            else reader.location
                        )
                    ],
                    language=Language.PYTHON,
                    p_url=p_url,
                    metadata=PythonRequirementsEntry(
                        name=package["name"],
                        extras=[],
                        markers=p_url,
                    ),
                    licenses=[],
                    type=PackageType.PythonPkg,
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": reader.location.path(),
                    }
                },
            )
            continue
    for package in toml_content.get("package", []):
        _pkg = next(
            (pkg for pkg in packages if pkg.name == package["name"]), None
        )
        dependencies: list[str] = list(package.get("dependencies", {}).keys())
        if _pkg and dependencies:
            pkg_dependencies = [
                pkg
                for dep in dependencies
                for pkg in packages
                if pkg.name == dep
            ]
            for dep in pkg_dependencies:
                relationships.append(
                    Relationship(
                        from_=dep,
                        to_=_pkg,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data=None,
                    )
                )

    return packages, relationships



from enum import (
    IntEnum,
)

# ref. https://github.com/rpm-software-management/rpm/blob
# /rpm-4.14.3-release/lib/rpmtag.h#L34
RPMTAG_HEADERIMAGE = 61
RPMTAG_HEADERSIGNATURES = 62
RPMTAG_HEADERIMMUTABLE = 63
HEADER_I18NTABLE = 100
RPMTAG_HEADERI18NTABLE = HEADER_I18NTABLE

# rpmTag_e
# ref. https://github.com/rpm-software-management/rpm
# /blob/rpm-4.14.3-release/lib/rpmtag.h#L34
RPMTAG_SIGMD5 = 261  # x
RPMTAG_NAME = 1000  # s
RPMTAG_VERSION = 1001  # s
RPMTAG_RELEASE = 1002  # s
RPMTAG_EPOCH = 1003  # i
RPMTAG_INSTALLTIME = 1008  # i
RPMTAG_SIZE = 1009  # i
RPMTAG_VENDOR = 1011  # s
RPMTAG_LICENSE = 1014  # s
RPMTAG_ARCH = 1022  # s
RPMTAG_FILESIZES = 1028  # i[] */
RPMTAG_FILEMODES = 1030  # h[] , specifically []uint16
# (ref https://github.com/rpm-software-management/rpm/blob
# /2153fa4ae51a84547129b8ebb3bb396e1737020e/lib/rpmtypes.h#L53 )*/
RPMTAG_FILEDIGESTS = 1035  # s[] */
RPMTAG_FILEFLAGS = 1037  # i[] */
RPMTAG_FILEUSERNAME = 1039  # s[] */
RPMTAG_FILEGROUPNAME = 1040  # s[] */
RPMTAG_SOURCERPM = 1044  # s */
RPMTAG_PROVIDENAME = 1047  # s[] */
RPMTAG_REQUIRENAME = 1049  # s[] */
RPMTAG_DIRINDEXES = 1116  # i[] */
RPMTAG_BASENAMES = 1117  # s[] */
RPMTAG_DIRNAMES = 1118  # s[] */
RPMTAG_FILEDIGESTALGO = 5011  # i  */
RPMTAG_SUMMARY = 1004  # s */
RPMTAG_PGP = 259  # b */

# rpmTag_enhances
# https://github.com/rpm-software-management/rpm/blob
# /rpm-4.16.0-release/lib/rpmtag.h#L375
RPMTAG_MODULARITYLABEL = 5096

# rpmTagType_e
# ref. https://github.com/rpm-software-management/rpm/blob
# /rpm-4.14.3-release/lib/rpmtag.h#L431
RPM_MIN_TYPE = 0
RPM_NULL_TYPE = 0
RPM_CHAR_TYPE = 1
RPM_INT8_TYPE = 2
RPM_INT16_TYPE = 3
RPM_INT32_TYPE = 4
RPM_INT64_TYPE = 5
RPM_STRING_TYPE = 6
RPM_BIN_TYPE = 7
RPM_STRING_ARRAY_TYPE = 8
RPM_I18NSTRING_TYPE = 9
RPM_MAX_TYPE = 9


class RpmTag(IntEnum):
    HEADERIMAGE = 61
    HEADERSIGNATURES = 62
    HEADERIMMUTABLE = 63
    HEADERI18NTABLE = 100
    SIGMD5 = 261
    NAME = 1000
    VERSION = 1001
    RELEASE = 1002
    EPOCH = 1003
    INSTALLTIME = 1008
    SIZE = 1009
    VENDOR = 1011
    LICENSE = 1014
    ARCH = 1022
    FILESIZES = 1028
    FILEMODES = 1030
    FILEDIGESTS = 1035
    FILEFLAGS = 1037
    FILEUSERNAME = 1039
    FILEGROUPNAME = 1040
    SOURCERPM = 1044
    PROVIDENAME = 1047
    REQUIRENAME = 1049
    DIRINDEXES = 1116
    BASENAMES = 1117
    DIRNAMES = 1118
    FILEDIGESTALGO = 5011
    SUMMARY = 1004
    PGP = 259
    MODULARITYLABEL = 5096


class RpmTagType(IntEnum):
    MIN_TYPE = 0
    NULL_TYPE = 0
    CHAR_TYPE = 1
    INT8_TYPE = 2
    INT16_TYPE = 3
    INT32_TYPE = 4
    INT64_TYPE = 5
    STRING_TYPE = 6
    BIN_TYPE = 7
    STRING_ARRAY_TYPE = 8
    I18NSTRING_TYPE = 9
    MAX_TYPE = 9



from fluid_sbom.linux.release import (
    Release,
)

PURL_QUALIFIER_ARCH = "arch"
PURL_QUALIFIER_DISTRO = "distro"
PURL_QUALIFIER_EPOCH = "epoch"
PURL_QUALIFIER_VCS_URL = "vcs_url"
PURL_QUALIFIER_UPSTREAM = "upstream"
PURL_CARGO_PKG_TYPE = "cargo"
PURL_GRADLE_PKG_TYPE = "gradle"


def purl_qualifiers(
    qualifiers: dict[str, str | None], release: Release | None = None
) -> dict[str, str]:
    # Handling distro qualifiers
    if release:
        distro_qualifiers = []
        if release.id_:
            distro_qualifiers.append(release.id_)
        if release.version_id:
            distro_qualifiers.append(release.version_id)
        elif release.build_id:
            distro_qualifiers.append(release.build_id)

        if distro_qualifiers:
            qualifiers[PURL_QUALIFIER_DISTRO] = "-".join(distro_qualifiers)

    return {
        key: qualifiers.get(key, "") or ""
        for key in sorted(qualifiers.keys())
        if qualifiers.get(key)
    }



from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.file_resolver.directory import (
    Directory,
)
import logging
import os
from pydantic import (
    BaseModel,
)

LOGGER = logging.getLogger(__name__)


class DirectoryConfig(BaseModel):
    path: str
    exclude: list[str] | None


class DirectorySource(BaseModel):
    config: DirectoryConfig
    resolver: Directory | None = None

    def file_resolver(self) -> Resolver:
        if self.resolver is None:
            self.resolver = Directory(root=self.config.path)
        return self.resolver


def new_from_directory_path(path: str) -> DirectorySource | None:
    return new_from_directory(DirectoryConfig(path=path, exclude=None))


def new_from_directory(cfg: DirectoryConfig) -> DirectorySource | None:
    if not os.path.isdir(cfg.path):
        LOGGER.error("Given path is not a directory: %s", cfg.path)
        return None

    return DirectorySource(config=cfg, resolver=None)



from fluid_sbom.file.coordinates import (
    Coordinates,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.metadata import (
    Metadata,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.file.type import (
    Type,
)
import glob
import magic
import os
from pydantic import (
    BaseModel,
)
from typing import (
    Callable,
    Generator,
    TextIO,
)


class Directory(Resolver, BaseModel):
    root: str

    def __post__init__(self, root: str) -> None:
        self.root = os.path.realpath(os.path.abspath(root))

    def has_path(self, path: str) -> bool:
        path = os.path.join(self.root, path.lstrip("/"))
        return os.path.exists(path)

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []
        for path in paths:
            relative_path = path.replace(self.root, "").lstrip("/")
            full_path = os.path.join(self.root, relative_path)
            if os.path.exists(full_path):
                locations.append(
                    Location(
                        coordinates=Coordinates(
                            real_path=full_path, file_system_id=""
                        ),
                        access_path=relative_path,
                        annotations={},
                    )
                )
        return locations

    def files_by_glob(self, *patters: str) -> list[Location]:
        result = []
        for pattern in patters:
            for item in glob.glob(pattern, root_dir=self.root):
                result.append(
                    Location(
                        coordinates=Coordinates(
                            real_path=os.path.join(self.root, item),
                            file_system_id="",
                        ),
                        access_path=item,
                        annotations={},
                    )
                )

        return result

    def files_by_mime_type(self, mime_type: str) -> list[Location]:
        matching_files = []
        mime_detector = magic.Magic(mime=True)

        for dirpath, _, filenames in os.walk(self.root):
            for filename in filenames:
                relative_path = (
                    os.path.join(dirpath, filename)
                    .replace(self.root, "")
                    .lstrip("/")
                )
                result_mime_type = mime_detector.from_file(relative_path)
                if mime_type == result_mime_type:
                    matching_files.append(
                        Location(
                            coordinates=Coordinates(
                                real_path=os.path.join(
                                    self.root, relative_path
                                ),
                                file_system_id="",
                            ),
                            access_path=relative_path,
                            annotations={},
                        )
                    )

        return matching_files

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: Callable[..., TextIO] | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        if (
            location.coordinates
            and location.coordinates.real_path is not None
            and os.path.exists(location.coordinates.real_path)
        ):
            return (function_reader or open)(  # type: ignore
                location.coordinates.real_path,
                encoding="utf-8",
                mode=mode or "r",
            )

        return None

    def file_metadata_by_location(self, location: Location) -> Metadata | None:
        link_destination = None
        if not location.access_path:
            return None

        stats = os.stat(location.access_path, follow_symlinks=False)

        if os.path.islink(location.access_path):
            file_type = Type.TYPE_SYM_LINK
            link_destination = os.readlink(location.access_path)
        elif os.path.isdir(location.access_path):
            file_type = Type.TYPE_DIRECTORY
        elif os.path.isfile(location.access_path):
            file_type = Type.TYPE_REGULAR

        mime_type = magic.Magic(mime=True).from_file(location.access_path)

        return Metadata(
            path=location.access_path,
            link_destination=link_destination or "",
            user_id=stats.st_uid,
            group_id=stats.st_gid,
            type=file_type,
            mime_type=mime_type,
        )

    def relative_file_path(self, _: Location, _path: str) -> Location:
        return Location(
            coordinates=Coordinates(
                real_path=_path, file_system_id="", line=None
            ),
            access_path=_path.replace(self.root, "").lstrip("/"),
        )

    def walk_file(self) -> Generator[str, None, None]:
        exclusions = {
            "node_modules",
            "dist",
            "__pycache__",
        }
        for dirpath, dirnames, filenames in os.walk(self.root):
            dirnames[:] = [dir for dir in dirnames if dir not in exclusions]

            for filename in filenames:
                full_path = os.path.join(dirpath, filename)
                relative_path = full_path.replace(self.root, "").lstrip("/")
                yield relative_path



from concurrent.futures import (
    ThreadPoolExecutor,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.debian.package import (
    new_dpkg_package,
)
from fluid_sbom.pkg.cataloger.debian.parse_dpkg_info_files import (
    parse_dpkg_conffile_info,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.dpkg import (
    DpkgDBEntry,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.file import (
    parse_bytes,
)
from multiprocessing import (
    cpu_count,
)
from pydantic import (
    BaseModel,
    ConfigDict,
)
import re
from typing import (
    cast,
    TextIO,
)


class DpkgField(BaseModel):
    name: str
    value: str
    model_config = ConfigDict(frozen=True)


def _split_deps(value: str) -> list[str]:
    fields = value.split(",")
    return [field.strip() for field in fields if field.strip()]


def extract_source_version(source: str) -> tuple[str, str | None]:
    if match_result := re.compile(
        r"(?P<name>\S+)( \((?P<version>.*)\))?"
    ).match(source):
        result = match_result.groupdict()
        return result["name"], result["version"] or ""
    return source, ""


def handle_new_key_value(line: str) -> tuple[str | None, str]:
    if ":" in line and not line.startswith(" "):
        key, value = line.split(":", 1)
        value = value.strip()
        match key:
            case "Installed-Size":
                value = str(parse_bytes(value))
        return key, value

    return None, line


def parse_dpkg_status(reader: TextIO) -> list[DpkgDBEntry]:
    content = reader.read()
    entries = []
    for package in content.strip().split("\n\n"):
        lines = package.split("\n")
        data = {}
        prev_key = ""
        for line in lines:
            key, value = handle_new_key_value(line)
            if key is not None:
                data[key] = value
                prev_key = key
            elif prev_key in data:
                data[prev_key] = f"{data[prev_key]}\n{value}"
        if all(not value for value in data.values()):
            continue
        source_name, source_version = extract_source_version(
            data.get("Source", "")
        )
        entry = DpkgDBEntry(
            package=data.get("Package", ""),
            source=source_name,
            version=data.get("Version", ""),
            source_version=source_version or "",
            architecture=data.get("Architecture", ""),
            maintainer=data.get("Maintainer", ""),
            installed_size=int(data.get("Installed-Size", "0")),
            description=data.get("Description", "").strip(),
            provides=_split_deps(data.get("Provides", "")),
            dependencies=_split_deps(data.get("Depends", "")),
            pre_dependencies=_split_deps(data.get("Pre-Depends", "")),
            files=parse_dpkg_conffile_info(data.get("Conffiles", "")),
        )
        entries.append(entry)

    return entries


def parse_dpkg_db(
    resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    entries: list[DpkgDBEntry] = parse_dpkg_status(reader.read_closer)
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        result = [
            pkg if isinstance(pkg, Package) else (pkg[0], pkg[1])
            for pkg in filter(
                None,
                executor.map(
                    lambda entry: new_dpkg_package(
                        entry,
                        reader.location,
                        resolver,
                        _env.linux_release if _env else None,
                    ),
                    entries,
                ),
            )
            if isinstance(pkg, (Package, tuple))
        ]
        packages = [
            item
            for pkg in result
            for item in (pkg if isinstance(pkg, tuple) else [pkg])
        ]
    return packages, associate_relationships(packages)


def associate_relationships(  # NOSONAR
    pkgs: list[Package],
) -> list[Relationship]:
    lookup: dict[str, list[Package]] = {}
    relationships: list[Relationship] = []
    for pkg in pkgs:
        lookup.setdefault(pkg.name, []).append(pkg)
    for pkg in pkgs:
        metadata = cast(DpkgDBEntry, pkg.metadata)
        for provides in metadata.provides or []:
            key = strip_version_specifier(provides)
            lookup.setdefault(key, []).append(pkg)

    for pkg in pkgs:
        metadata = cast(DpkgDBEntry, pkg.metadata)
        all_deps = [
            *(metadata.dependencies or []),
            *(metadata.pre_dependencies or []),
        ]
        for dep_specifier in all_deps:
            deps = split_package_choices(dep_specifier)
            for dep in deps:
                for dep_pkg in lookup.get(dep, []):
                    relationships.append(
                        Relationship(
                            from_=dep_pkg,
                            to_=pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                            data=None,
                        )
                    )

    return relationships


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next(
        (i for i, char in enumerate(item) if char in specifiers), None
    )

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def split_package_choices(value: str) -> list[str]:
    fields = value.split("|")
    return [
        strip_version_specifier(field) for field in fields if field.strip()
    ]



from fluid_sbom.internal.cache import (
    dual_cache,
)
from retry_requests import (  # type: ignore
    retry,
)
from typing import (
    Any,
)


@dual_cache
def make_get(url: str, *, content: bool = False, **kwargs: Any) -> Any | None:
    response = retry().get(url, timeout=kwargs.pop("timeout", 20), **kwargs)
    if response.status_code != 200:
        return None
    if content:
        return response.content.decode("utf-8")

    return response.json()



from fluid_sbom.utils.file import (
    Digest,
)
import hashlib
from io import (
    BufferedReader,
)


def new_digests_from_file(
    file_object: BufferedReader, hashes: list[str]
) -> list[Digest]:
    # Create hash objects
    hash_objects = [
        hashlib.new(hash_key, usedforsecurity=False) for hash_key in hashes
    ]

    # Read file and update each hash object
    while chunk := file_object.read(4096):
        for hasher in hash_objects:
            hasher.update(chunk)

    # Prepare the result list of Digest objects
    if file_object.tell() == 0:  # Check if file size is zero
        return []

    return [
        Digest(
            algorithm=hash_name,
            value=hasher.hexdigest(),
        )
        for hash_name, hasher in zip(hashes, hash_objects)
    ]



from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.constants import (
    HASH_INDEX_ENTRY_SIZE,
    HASH_OF_INDEX_PAGE_TYPE,
    HASH_OFF_PAGE_SIZE,
    OVERFLOW_PAGE_TYPE,
    PAGE_HEADER_SIZE,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.hash_page_entry import (
    parse_hash_off_page_entry,
)
import io
from pydantic import (
    BaseModel,
)
import struct


class HashPage(BaseModel):
    """
    Represents a Hash Page in the RPM database.
    """

    lsn: bytes  # 8 bytes
    page_no: int  # uint32
    previous_page_no: int  # uint32
    next_page_no: int  # uint32
    num_entries: int  # uint16
    free_area_offset: int  # uint16
    tree_level: int  # uint8
    page_type: int  # uint8

    @classmethod
    def from_bytes(  # pylint:disable=too-many-locals
        cls,
        data: bytes,
        swapped: bool,
    ) -> "HashPage":
        expected_size = 26  # Total size in bytes for HashPage header
        if len(data) < expected_size:
            raise ValueError(
                "Data too short, expected at least "
                f"{expected_size} bytes, got {len(data)} bytes"
            )

        byte_order_char = ">" if swapped else "<"

        fmt = f"{byte_order_char}8sIIIHHBB"
        try:
            unpacked_data = struct.unpack(fmt, data[:expected_size])
        except struct.error as exc:
            raise ValueError("Failed to unpack HashPage") from exc

        (
            lsn,
            page_no,
            previous_page_no,
            next_page_no,
            num_entries,
            free_area_offset,
            tree_level,
            page_type,
        ) = unpacked_data

        return cls(
            lsn=lsn,
            page_no=page_no,
            previous_page_no=previous_page_no,
            next_page_no=next_page_no,
            num_entries=num_entries,
            free_area_offset=free_area_offset,
            tree_level=tree_level,
            page_type=page_type,
        )

    def __repr__(self) -> str:
        return (
            f"HashPage(LSN={self.lsn.hex()}, PageNo={self.page_no},"
            f" PreviousPageNo={self.previous_page_no},"
            f" NextPageNo={self.next_page_no}, NumEntries={self.num_entries},"
            f" FreeAreaOffset={self.free_area_offset},"
            f" TreeLevel={self.tree_level}, PageType={self.page_type})"
        )


def parse_hash_page(data: bytes, swapped: bool) -> HashPage:
    return HashPage.from_bytes(data, swapped)


def hash_page_value_content(
    db_file: io.BufferedReader,
    page_data: bytes,
    hash_page_index: int,
    page_size: int,
    swapped: bool,
) -> bytes:
    # The first byte is the page type
    value_page_type = page_data[hash_page_index]

    # Only HOFFPAGE page types have data of interest
    if value_page_type != HASH_OF_INDEX_PAGE_TYPE:
        raise ValueError(
            f"Only HOFFPAGE types supported (got {value_page_type})"
        )

    hash_off_page_entry_buff = page_data[
        hash_page_index : hash_page_index + HASH_OFF_PAGE_SIZE
    ]

    entry = parse_hash_off_page_entry(hash_off_page_entry_buff, swapped)

    hash_value = bytearray()

    current_page_no = entry.page_no
    while current_page_no != 0:
        page_start = page_size * current_page_no

        db_file.seek(page_start)
        current_page_buff = db_file.read(page_size)
        if len(current_page_buff) != page_size:
            raise ValueError(
                f"Failed to read page {current_page_no}: insufficient data"
            )

        current_page = parse_hash_page(current_page_buff, swapped)
        if current_page.page_type != OVERFLOW_PAGE_TYPE:
            current_page_no = current_page.next_page_no
            continue

        if current_page.next_page_no == 0:
            # This is the last page, the content length is FreeAreaOffset
            data_length = current_page.free_area_offset
            hash_value_bytes = current_page_buff[
                PAGE_HEADER_SIZE : PAGE_HEADER_SIZE + data_length
            ]
        else:
            # For intermediate pages, use all data after header
            hash_value_bytes = current_page_buff[PAGE_HEADER_SIZE:]

        hash_value.extend(hash_value_bytes)

        current_page_no = current_page.next_page_no

    return bytes(hash_value)


def hash_page_value_indexes(data: bytes, entries: int, swapped: bool) -> list:
    if entries % 2 != 0:
        raise ValueError(
            "Invalid hash index: entries should only come in"
            f" pairs (got {entries})"
        )

    order_char = ">" if swapped else "<"

    hash_index_values = []

    # Every entry is a 2-byte offset that points somewhere in the current
    # database page.
    hash_index_size = entries * HASH_INDEX_ENTRY_SIZE
    hash_index_data = data[
        PAGE_HEADER_SIZE : PAGE_HEADER_SIZE + hash_index_size
    ]

    # Data is stored in key-value pairs, skip over keys and only keep values
    key_value_pair_size = 2 * HASH_INDEX_ENTRY_SIZE

    for idx in range(0, len(hash_index_data), HASH_INDEX_ENTRY_SIZE):
        if (idx - HASH_INDEX_ENTRY_SIZE) % key_value_pair_size == 0:
            # This is a value index
            value_bytes = hash_index_data[idx : idx + HASH_INDEX_ENTRY_SIZE]
            if len(value_bytes) < HASH_INDEX_ENTRY_SIZE:
                raise ValueError(f"Insufficient data at index {idx}")
            value = struct.unpack(f"{order_char}H", value_bytes)[0]
            hash_index_values.append(value)

    return hash_index_values


def slice_reader(reader: io.BufferedReader, num: int) -> bytes:
    """
    Reads n bytes from a reader.

    :param reader: A file-like object to read from.
    :param n: Number of bytes to read.
    :return: The bytes read.
    :raises ValueError: If the number of bytes read is less than n.
    """
    data = reader.read(num)
    if len(data) != num:
        raise ValueError(f"Failed to read {num} bytes, got {len(data)} bytes")
    return data



from os import (
    environ,
)
from typing import (
    Literal,
)


def guess_environment() -> Literal["development"] | Literal["production"]:
    return (
        "production"
        if environ.get("CI_COMMIT_REF_NAME", "trunk") == "trunk"
        else "development"
    )



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.java.package import (
    package_url,
)
from fluid_sbom.pkg.java import (
    JavaArchive,
    JavaPomProperties,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
import os
from pathlib import (
    Path,
)
from pydantic import (
    ValidationError,
)
import tempfile
import zipfile

LOGGER = logging.getLogger(__name__)


def parse_apk(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    with tempfile.TemporaryDirectory() as output_folder:
        with zipfile.ZipFile(reader.read_closer.name, "r") as apk_file:
            apk_file.extractall(output_folder)
        files: list[str] = []
        meta_dir = os.path.join(output_folder, "META-INF")
        if os.path.exists(meta_dir):
            files = [
                os.path.join(meta_dir, file)
                for file in os.listdir(meta_dir)
                if file.endswith(".version")
            ]
        for file in files:
            with open(file, "r", encoding="utf-8") as version_reader:
                version = version_reader.read().strip()
            parts = Path(file).name.replace(".version", "").split("_", 1)
            group_id = parts[0]
            if not group_id:
                raise ValueError(
                    f"Invalid group_id for {file} in {reader.location}"
                )
            artifact_id = parts[1]
            java_archive = JavaArchive(
                pom_properties=JavaPomProperties(
                    group_id=group_id,
                    artifact_id=artifact_id,
                    version=version,
                )
            )
            try:
                packages.append(
                    Package(
                        name=artifact_id,
                        version=version,
                        licenses=[],
                        locations=[reader.location],
                        language=Language.JAVA,
                        type=PackageType.JavaPkg,
                        metadata=java_archive,
                        p_url=package_url(artifact_id, version, java_archive),
                    )
                )
            except ValidationError as ex:
                LOGGER.warning(
                    "Malformed package. Required fields are missing or data "
                    "types are incorrect.",
                    extra={
                        "extra": {
                            "exception": ex.errors(include_url=False),
                            "location": reader.location.path(),
                        }
                    },
                )
                continue
    return packages, []



from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.file_resolver import (
    gen_location,
)
from fluid_sbom.linux.release import (
    identify_release,
)
from fluid_sbom.pkg.cataloger.generic.cataloger import (
    execute_parsers,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.handle import (
    handle_parser,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.python import (
    PythonPackage,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedException,
)
from fluid_sbom.utils.package import (
    strip_version_specifier,
)
import logging
import multiprocessing
import reactivex
import reactivex.operators
from reactivex.scheduler import (
    ThreadPoolScheduler,
)
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def handle_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package in packages:
        match package.found_by:
            case "python-installed-package-cataloger":
                python_package: PythonPackage = cast(
                    PythonPackage, package.metadata
                )
                for dep in (
                    python_package.dependencies
                    if python_package.dependencies
                    else []
                ):
                    dep_name = strip_version_specifier(dep)
                    if dep_package := next(
                        (x for x in packages if x.name == dep_name), None
                    ):
                        relationships.append(
                            Relationship(
                                from_=dep_package,
                                to_=package,
                                type=(
                                    RelationshipType.DEPENDENCY_OF_RELATIONSHIP
                                ),
                                data=None,
                            )
                        )
    return relationships


def package_operations_factory(
    resolver: Resolver,
) -> tuple[list[Package], list[Relationship]]:
    observer = reactivex.from_iterable(resolver.walk_file())
    result_packages: list[Package] = []
    result_relations: list[Relationship] = []
    completed_event = multiprocessing.Event()
    errors = []

    def on_completed() -> None:
        completed_event.set()

    def on_error(error: Exception) -> None:
        errors.append(error)
        on_completed()

    def on_next(value: tuple[list[Package], list[Relationship]]) -> None:
        packages, relations = value
        result_packages.extend(packages)
        result_relations.extend(relations)

    optimal_thread_count = multiprocessing.cpu_count()
    pool_scheduler = ThreadPoolScheduler(optimal_thread_count)

    observer.pipe(
        handle_parser(scheduler=pool_scheduler),
        gen_location(resolver),
        execute_parsers(
            resolver, Environment(linux_release=identify_release(resolver))
        ),
    ).subscribe(
        on_next,
        on_error=on_error,
        on_completed=on_completed,
        scheduler=pool_scheduler,
    )
    completed_event.wait()
    result_relations.extend(handle_relationships(result_packages))

    if errors:
        for error in errors:
            raise UnexpectedException(error) from error

    return result_packages, result_relations



from fluid_sbom.internal.cache import (
    dual_cache,
)
import requests
from typing import (
    Any,
)


@dual_cache
def get_gem_package(
    package_name: str, version: str | None = None
) -> dict[str, Any] | None:
    if version:
        url = (
            "https://rubygems.org/api/v2/rubygems/"
            f"{package_name}/versions/{version}.json"
        )
    else:
        url = f"https://rubygems.org/api/v1/gems/{package_name}.json"

    response = requests.get(url, timeout=20)
    if response.status_code == 200:
        package_info = response.json()
        return package_info

    return None



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.python.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from typing import (
    cast,
    ItemsView,
)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: IndexedDict) -> str:
    version: str = value["version"]
    return version.strip("=<>~^ ")


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict | None,
) -> list[Package]:
    if dependencies is None:
        return []

    items: ItemsView[str, IndexedDict] = dependencies.items()
    return [
        Package(
            name=name,
            version=_get_version(value),
            locations=[
                _get_location(reader.location, value.position.start.line)
            ],
            language=Language.PYTHON,
            type=PackageType.PythonPkg,
            p_url=package_url(
                name=name,
                version=_get_version(value),
                package=None,
            ),
            licenses=[],
        )
        for name, value in items
    ]


def parse_pipfile_lock_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
    )
    deps: IndexedDict | None = content.get("default")
    dev_deps: IndexedDict | None = content.get("develop")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps),
    ]
    return packages, []



from fluid_sbom.config.logger import (
    LOGGER,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.errors import (
    InvalidDBFormat,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)
import sqlite3
from typing import (
    Generator,
)


class Sqlite(RpmDBInterface):  # pylint: disable=too-few-public-methods
    def __init__(self, connection: sqlite3.Connection) -> None:
        self.connection = connection
        super().__init__()

    def read(
        self,
    ) -> Generator[bytes, None, None]:
        cursor = self.connection.cursor()
        try:
            blobs = cursor.execute("SELECT blob FROM Packages;").fetchall()
        except sqlite3.DatabaseError:
            return

        for blob in blobs:
            yield blob[0]


def open_sqlite(file_path: str) -> RpmDBInterface:
    connection = sqlite3.connect(file_path)
    cursor = connection.cursor()
    try:
        cursor.execute("PRAGMA schema_version;")
    except sqlite3.DatabaseError as exc:
        LOGGER.warning(
            "Invalid SQLite database file",
            extra={
                "extra": {
                    "location": file_path,
                }
            },
        )
        raise InvalidDBFormat from exc
    return Sqlite(connection)



from cyclonedx.validation import (
    ValidationError as CycloneDXError,
)
from jsonschema import (
    ValidationError as JSONSchemaError,
)
from spdx_tools.spdx.validation.validation_message import (
    ValidationMessage,
)
from tree_sitter import (
    Node,
)


class CustomBaseException(Exception):
    """Base exception class for custom exceptions."""


class CycloneDXValidationError(CustomBaseException):
    """Exception for CycloneDX validation errors."""

    def __init__(self, error: CycloneDXError) -> None:
        """Constructor"""
        header = "Exception - CycloneDx validation error\n"
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class DuplicatedKey(Exception):
    """Exception raised for duplicated keys."""


class ForbiddenModuleImported(Exception):
    """Exception raised when a forbidden module is imported."""


class FluidJSONValidationError(CustomBaseException):
    """Exception for Fluid JSON validation errors."""

    def __init__(self, error: JSONSchemaError) -> None:
        """Constructor"""
        header = "Exception - Fluid JSON validation error\n"
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class InvalidDocumentKeyPair(Exception):
    """Exception raised for invalid document key pairs."""


class InvalidType(Exception):
    """Exception raised for invalid types."""


class SPDXValidationError(CustomBaseException):
    """Exception for SPDX validation errors."""

    def __init__(self, error_messages: list[ValidationMessage]) -> None:
        """Constructor"""
        header = "Exception - SPDX validation error\n"
        error_details = "\n".join(
            (
                f"Validation error: {message.validation_message}\n"
                f"Context: {message.context}"
            )
            for message in error_messages
        )
        msg = f"❌ {header}{error_details}"
        super().__init__(msg)


class UnexpectedException(CustomBaseException):
    """Exception for unexpected errors encountered during SBOM execution."""

    def __init__(self, error: Exception) -> None:
        """Constructor"""
        header = (
            "Exception - An unexpected exception was encountered "
            "during SBOM execution. The process will be terminated to prevent "
            "potential inconsistencies.\n"
        )
        msg = f"❌ {header}{error}"
        super().__init__(msg)


class UnexpectedNode(Exception):
    """Exception raised for unexpected nodes."""

    def __init__(self, node: Node | str) -> None:
        type_name = node.type if isinstance(node, Node) else node
        value = node.text.decode("utf-8") if isinstance(node, Node) else None
        if value:
            super().__init__(
                f"Unexpected node type {type_name} with value {value}"
            )
        else:
            super().__init__(f"Unexpected node type {type_name}")


class UnexpectedNodeType(Exception):
    """Exception raised for unexpected node types."""

    def __init__(
        self, node: Node | str, expected_type: str | None = None
    ) -> None:
        type_name = node.type if isinstance(node, Node) else node
        if expected_type:
            super().__init__(
                f"Unexpected node type {type_name} for {expected_type}"
            )
        else:
            super().__init__(f"Unexpected node type {type_name}")


class UnexpectedValueType(Exception):
    """Exception raised for unexpected value types."""


class UnexpectedChildrenLength(Exception):
    """Exception raised for nodes with unexpected number of children."""

    def __init__(self, node: Node | str, expected_length: int) -> None:
        type_name = node.type if isinstance(node, Node) else node
        super().__init__(
            f"Unexpected node type {type_name} for {expected_length} children"
        )



from contextlib import (
    suppress,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
    IndexedList,
)
from fluid_sbom.utils.exceptions import (
    DuplicatedKey,
    InvalidDocumentKeyPair,
    InvalidType,
    UnexpectedNode,
)
from more_itertools import (
    mark_ends,
)
import os
import re
from tree_sitter import (
    Language,
    Node,
    Parser,
)
from typing import (
    Any,
    Iterator,
)


def _validate_key_pairs(root_node: Node) -> None:
    all_children_pairs = tuple(
        node.start_point[0]
        for node in root_node.children
        if node.type == "pair"
    )
    if len(all_children_pairs) != len(set(all_children_pairs)):
        raise InvalidDocumentKeyPair()


def handle_string(node: Node) -> str:
    value = node.text.decode("utf-8")
    str_quote = "'" if value.startswith("'") else None
    str_quote = str_quote or ('"' if value.startswith('"') else None)
    if (
        str_quote
        and (value.startswith(str_quote))
        and (value.endswith(str_quote))
    ):
        value = value.strip(str_quote)
        # Remove surrounding triple quotes if present
        value = re.sub(r'"""|\'\'\'', "", value)
        # Replace multiple spaces and newlines with a single space
        # value = re.sub(r"\s+", " ", value)
        # Trim leading and trailing spaces
        value = value.strip()

    return value


def handle_array(node: Node) -> IndexedList:
    data = IndexedList(node)
    for children_node in node.children:
        if children_node.type in ("[", "]", ",", "comment"):
            continue
        data.append((handle_node(children_node), children_node))
    return data


def handle_bare_key(node: Node) -> str:
    return handle_string(node)


def handle_quoted_key(node: Node) -> str:
    value = handle_string(node)
    if value.startswith('"') or value.startswith("'"):
        return value[1:-1]
    return value


def handle_dotted_key(node: Node) -> list[str]:
    value = handle_string(node)
    parts = re.findall(r'"[^"]*"|[\w-]+', value)
    parts = [part.strip('" ') for part in parts]
    return parts


def handle_table_array_element(
    node: Node,
) -> tuple[tuple[str | list[str], Node], IndexedDict]:
    data = IndexedDict(node)
    bare_key = node.named_children[0]
    bare_key_value = handle_node(bare_key)

    # the children's are pair nods
    for children_node in node.named_children[1:]:
        if children_node.type in ("[", "]", ", ", "comment"):
            continue
        if children_node.type != "pair":
            raise UnexpectedNode(children_node)
        key, value = handle_node(children_node)
        data[key] = value
    return (bare_key_value, bare_key), data


def handle_pair(node: Node) -> tuple[tuple[str, Node], tuple[Any, Node]]:
    bare_key_node, value_node = node.named_children[:2]
    bare_key_value = handle_node(bare_key_node)
    value_resolved = handle_node(value_node)
    return (bare_key_value, bare_key_node), (value_resolved, value_node)


def handle_boolean(node: Node) -> bool:
    return node.text.decode("utf-8").lower() == "true"


def handle_integer(node: Node) -> int:
    decode_str = node.text.decode("utf-8")
    with suppress(ValueError):
        return int(decode_str)
    with suppress(ValueError):
        return int(decode_str, 16)
    with suppress(ValueError):
        return int(decode_str, 8)

    raise ValueError(f"Invalid integer value: {decode_str}")


def handle_float(node: Node) -> float:
    decoded_str = node.text.decode("utf-8")
    with suppress(ValueError):
        return float(decoded_str)

    with suppress(ValueError):
        return float(decoded_str.replace(".", "").lower())

    raise ValueError(f"Invalid float value: {decoded_str}")


def handle_local_date(node: Node) -> str:
    return handle_string(node)


def handle_inline_table(node: Node) -> IndexedDict:
    data = IndexedDict(node)
    for children_node in node.named_children:
        key, value = handle_node(children_node)
        nested_dict(
            data=data,
            keys=key[0],
            keys_node=key[1],
            value=value[0],
            value_node=value[1],
        )
    return data


def handle_table(node: Node, data_1: IndexedDict) -> None:
    data_new = IndexedDict(node)
    for children_node in node.named_children[1:]:
        if children_node.type == "comment":
            continue
        key, value = handle_node(children_node)
        nested_dict(
            data=data_new,
            keys=key[0],
            keys_node=key[1],
            value=value[0],
            value_node=value[1],
        )
    bare_key_node = node.named_children[0]
    bare_key_value = handle_node(bare_key_node)
    if isinstance(bare_key_value, list):
        nested_dict(
            data=data_1,
            keys=bare_key_value,
            keys_node=bare_key_node,
            value=data_new,
            value_node=node,
        )
    else:
        data_1[(bare_key_value, bare_key_node)] = (
            data_new,
            node,
        )


def handle_node(node: Node) -> Any:
    value: Any | None = None
    match node.type:
        case "comment":
            value = None
        case "string":
            value = handle_string(node)
        case "array":
            value = handle_array(node)
        case "pair":
            value = handle_pair(node)
        case "bare_key":
            value = handle_bare_key(node)
        case "quoted_key":
            value = handle_quoted_key(node)
        case "dotted_key":
            value = handle_dotted_key(node)
        case "boolean":
            value = handle_boolean(node)
        case "integer":
            value = handle_integer(node)
        case "float":
            value = handle_float(node)
        case "inline_table":
            value = handle_inline_table(node)
        case "local_date":
            value = handle_local_date(node)
        case _:
            raise UnexpectedNode(node.type)

    return value


def _nested_dict_handle_list(data: Any) -> IndexedDict:
    if not isinstance(data, IndexedDict):
        if not isinstance(data, IndexedList):
            raise InvalidType("attempted to extend non-table type")
        data = data[-1]
    return data


def _nested_dict_validate_duplicated_data(
    data: IndexedDict, key: str, is_last: bool
) -> None:
    if is_last and key in data:
        raise DuplicatedKey(f"Defining a key multiple times is invalid: {key}")


def nested_dict(
    *,
    data: IndexedDict,
    keys: list[str] | str,
    keys_node: Node,
    value: Any | None = None,
    value_node: Node | None = None,
) -> None:
    build_key = []
    for _, is_last, key in _mark_ends(keys):
        build_key.append(key)
        data = _nested_dict_handle_list(data)

        _nested_dict_validate_duplicated_data(data, key, is_last)
        if key not in data:
            if is_last:
                data[(key, keys_node)] = (
                    value if value is not None else IndexedDict(),
                    value_node or keys_node,
                )
                continue
            data[(key, keys_node)] = (IndexedDict(), value_node or keys_node)
        data = data[key]


def _mark_ends(keys: list[str] | str) -> Iterator[tuple[bool, bool, str]]:
    yield from mark_ends(keys if isinstance(keys, list) else [keys])


def _nested_list_handle_list(data: Any) -> IndexedDict:
    if isinstance(data, IndexedList):
        data = data[-1]
    return data


def _nested_list_handle_fists_keys(
    data: IndexedDict, key: str, is_last: bool
) -> bool:
    return key not in data and not is_last


def _nested_list_handle_last_keys(
    data: IndexedDict, key: str, is_last: bool
) -> bool:
    return key not in data and is_last


def _nested_list_handle_fists_keys_exists(
    data: IndexedDict, key: str, is_last: bool
) -> bool:
    return key in data and not is_last


def nested_list(
    *,
    data: IndexedDict,
    keys: list[str] | str,
    keys_node: Node,
    value: Any | None = None,
    value_node: Node | None = None,
) -> None:
    build_key = []
    for _, is_last, key in _mark_ends(keys):
        build_key.append(key)
        data = _nested_list_handle_list(data)

        if _nested_list_handle_fists_keys(data, key, is_last):
            data[(key, keys_node)] = (IndexedDict(), value_node or keys_node)
            data = data[key]
            continue
        if _nested_list_handle_fists_keys_exists(data, key, is_last):
            data = data[key]
            continue

        if _nested_list_handle_last_keys(data, key, is_last):
            data[(key, keys_node)] = (
                IndexedList(keys_node),
                value_node or keys_node,
            )
        if (
            is_last
        ):  # if the doted key is the last, the item is a value to append
            try:
                data[key].append((value, value_node or keys_node))
            except AttributeError as exc:
                raise InvalidType(
                    (
                        "attempted to extend non-table  "
                        f" type: {'.'.join(build_key)}"
                    )
                ) from exc
            continue
        data = data[key]


def parse_toml_with_tree_sitter(toml_content: str) -> IndexedDict:
    parser = Parser()
    parser.set_language(
        Language(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "toml.so",
            ),
            "toml",
        )
    )

    result = parser.parse(toml_content.encode("utf-8"))
    data: IndexedDict = IndexedDict(result.root_node)
    if result.root_node.type != "document":
        return data
    _validate_key_pairs(result.root_node)
    for node in result.root_node.children:
        if node.type == "pair":
            key, pair_value = handle_node(node)
            nested_dict(
                data=data,
                keys=key[0],
                keys_node=key[1],
                value=pair_value[0],
                value_node=pair_value[1],
            )
        elif node.type == "table":
            handle_table(node, data)
        elif node.type == "table_array_element":
            key, value = handle_table_array_element(node)
            nested_list(
                data=data,
                keys=key[0],
                keys_node=key[1],
                value=value,
                value_node=node,
            )
        elif node.type == "ERROR":
            raise UnexpectedNode(node)

    return data



from contextlib import (
    suppress,
)
from pydantic import (
    BaseModel,
    Field,
)
import tarfile

# IEC Sizes
BYTE = 1 << (0 * 10)
KIBYTE = 1 << (1 * 10)
MIBYTE = 1 << (2 * 10)
GIBYTE = 1 << (3 * 10)
TIBYTE = 1 << (4 * 10)
PIBYTE = 1 << (5 * 10)
EIBYTE = 1 << (6 * 10)

# SI Sizes
IBYTE = 1
KBYTE = IBYTE * 1000
MBYTE = KBYTE * 1000
GBYTE = MBYTE * 1000
TBYTE = GBYTE * 1000
PBYTE = TBYTE * 1000
EBYTE = PBYTE * 1000

bytes_size_table = {
    "b": BYTE,
    "kib": KIBYTE,
    "kb": KBYTE,
    "mib": MIBYTE,
    "mb": MBYTE,
    "gib": GIBYTE,
    "gb": GBYTE,
    "tib": TIBYTE,
    "tb": TBYTE,
    "pib": PIBYTE,
    "pb": PBYTE,
    "eib": EIBYTE,
    "eb": EBYTE,
    # Without suffix
    "": BYTE,
    "ki": KIBYTE,
    "k": KBYTE,
    "mi": MIBYTE,
    "m": MBYTE,
    "gi": GIBYTE,
    "g": GBYTE,
    "ti": TIBYTE,
    "t": TBYTE,
    "pi": PIBYTE,
    "p": PBYTE,
    "ei": EIBYTE,
    "e": EBYTE,
}


def parse_bytes(size_human: str) -> int:
    # Initialize the last digit index and the flag for comma presence
    last_digit = 0
    has_comma = False

    # Identify the last digit or comma in the string
    for i, char in enumerate(size_human):
        if not (char.isdigit() or char == "." or char == ","):
            break
        if char == ",":
            has_comma = True
        last_digit = i + 1

    # Extract the number part and remove commas if present
    num_str = size_human[:last_digit]
    if has_comma:
        num_str = num_str.replace(",", "")

    # Convert the number string to float
    try:
        num = int(num_str)
    except ValueError as exc:
        raise ValueError(f"Could not parse number: {num_str}") from exc

    # Get the unit and calculate the final value
    extra = size_human[last_digit:].strip().lower()
    if extra in bytes_size_table:
        num *= bytes_size_table[extra]
        return num

    raise ValueError(f"Unhandled size name: {extra}")


def extract_tar_file(tar_path: str, out_path: str) -> None:
    with tarfile.open(tar_path, "r") as tar:
        for member in tar.getmembers():
            member.name = member.name.lstrip("/")
            with suppress(Exception):
                tar.extract(member, path=out_path)


class Digest(BaseModel):
    algorithm: str | None = Field(min_length=1)
    value: str | None = Field(min_length=1)



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.swift.package import (
    new_cocoa_pods_package,
)
from fluid_sbom.pkg.package import (
    Package,
)
import logging
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def generate_relations(
    dependencies_index: dict[str, list[str]], packages: list[Package]
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package_name, dependencies in dependencies_index.items():
        package = next(x for x in packages if x.name == package_name)
        for dependency in dependencies:
            if package_dep := next(
                (x for x in packages if x.name == dependency.split(" ")[0]),
                None,
            ):
                relationships.append(
                    Relationship(
                        from_=package_dep,
                        to_=package,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data=None,
                    )
                )

    return relationships


def parse_podfile_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        podfile: IndexedDict = cast(
            IndexedDict, parse_yaml_with_tree_sitter(reader.read_closer.read())
        )
    except ValueError:
        LOGGER.warning("Failed to parse package.json: %s", reader.location)
        return [], []

    if not podfile or "PODS" not in podfile:
        return [], []

    packages: list[Package] = []
    relationships: list[Relationship] = []
    dependencies_index: dict[str, list[str]] = {}
    for index, pod in enumerate(podfile["PODS"]):
        new_location = deepcopy(reader.location)
        pod_blob = ""
        pod_name = ""
        if isinstance(pod, str):
            pod_blob = pod
        elif isinstance(pod, IndexedDict):
            pod_blob = list(pod.keys())[0]
            pod_name = pod_blob.split(" ")[0]
            dependencies_index[pod_name] = [
                x.split(" ")[0] for x in pod[pod_blob]
            ]

        pod_name = pod_blob.split(" ")[0]
        pod_version = pod_blob.split(" ")[1].strip("()")
        pod_root_package = pod_name.split("/")[0]

        if pod_root_package not in podfile["SPEC CHECKSUMS"]:
            LOGGER.error("Malformed podfile.lock. Incomplete checksums")
            return [], []

        if new_location.coordinates:
            new_location.coordinates.line = (
                podfile["PODS"].get_position(index).start.line
            )

        if pkg := new_cocoa_pods_package(
            pod_name,
            pod_version,
            podfile["SPEC CHECKSUMS"][pod_root_package],
            new_location,
        ):
            packages.append(pkg)

    relationships = generate_relations(dependencies_index, packages)

    return packages, relationships



from fluid_sbom.internal.package_information.ruby import (
    get_gem_package,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
from packageurl import (
    PackageURL,
)


def package_url(name: str, version: str) -> str:
    return PackageURL(
        type="gem",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def complete_package(package: Package) -> Package:
    current_package = get_gem_package(package.name, package.version)
    gem_package = get_gem_package(package.name)
    if not gem_package:
        return package

    package.health_metadata = HealthMetadata(
        latest_version=gem_package["version"],
        latest_version_created_at=gem_package["version_created_at"],
    )
    if current_package:
        digest_value = gem_package.get("sha") or None
        package.health_metadata.artifact = Artifact(
            url=gem_package["gem_uri"],
            integrity=Digest(
                algorithm="sha" if digest_value else None,
                value=digest_value,
            ),
        )

    package.health_metadata.authors = gem_package["authors"]
    if not package.licenses and gem_package["licenses"]:
        package.licenses = gem_package["licenses"]
    return package



import abc
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.metadata import (
    Metadata,
)
from pydantic import (
    BaseModel,
)
from typing import (
    Callable,
    Generator,
    TextIO,
    Type,
)


class ContentResolver(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(
        cls: Type["ContentResolver"], subclass: object
    ) -> bool:
        return (
            hasattr(subclass, "file_contents_by_location")
            and callable(subclass.file_contents_by_location)
            or NotImplemented
        )

    @abc.abstractmethod
    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: Callable[..., TextIO] | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        """Resolve file context"""
        raise NotImplementedError


class MetadataResolver(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(
        cls: Type["MetadataResolver"], subclass: object
    ) -> bool:
        return (
            hasattr(subclass, "file_metadata_by_location")
            and callable(subclass.file_metadata_by_location)
            or NotImplemented
        )

    @abc.abstractmethod
    def file_metadata_by_location(
        self,
        location: Location,
    ) -> Metadata | None:
        """Resolve file context"""
        raise NotImplementedError


class PathResolver(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls: Type["PathResolver"], subclass: object) -> bool:
        return (
            hasattr(subclass, "has_path")
            and callable(subclass.has_path)
            and hasattr(subclass, "files_by_path")
            and callable(subclass.files_by_path)
            and hasattr(subclass, "files_by_glob")
            and callable(subclass.files_by_glob)
            and hasattr(subclass, "files_by_mime_type")
            and callable(subclass.files_by_mime_type)
            and hasattr(subclass, "relative_file_path")
            and callable(subclass.relative_file_path)
            and hasattr(subclass, "walk_file")
            and callable(subclass.walk_file)
            or NotImplemented
        )

    @abc.abstractmethod
    def has_path(self, path: str) -> bool:
        """Resolve file context"""
        raise NotImplementedError

    @abc.abstractmethod
    def files_by_path(self, *paths: str) -> list[Location]:
        """Resolve file context"""
        raise NotImplementedError

    @abc.abstractmethod
    def files_by_glob(self, *patters: str) -> list[Location]:
        """Resolve file context"""
        raise NotImplementedError

    @abc.abstractmethod
    def files_by_mime_type(self, mime_type: str) -> list[Location]:
        """Resolve file context"""
        raise NotImplementedError

    @abc.abstractmethod
    def relative_file_path(self, _: Location, path: str) -> Location | None:
        """Resolve file context"""
        raise NotImplementedError

    @abc.abstractmethod
    def walk_file(self) -> Generator[str, None, None]:
        raise NotImplementedError


class Resolver(
    ContentResolver,
    PathResolver,
    MetadataResolver,
    BaseModel,
    metaclass=abc.ABCMeta,
):
    pass



from bs4 import (
    BeautifulSoup,
)
from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def parse_dotnet_pkgs_config(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    packages = []

    for pkg in root.find_all("package", recursive=True):
        if (id_ := pkg.get("id")) and (version := pkg.get("version")):
            line = pkg.sourceline
            location = deepcopy(reader.location)
            if location.coordinates:
                location.coordinates.line = line
            try:
                packages.append(
                    Package(
                        name=id_,
                        version=version,
                        locations=[location],
                        language=Language.DOTNET,
                        licenses=[],
                        type=PackageType.DotnetPkg,
                        metadata=None,
                        p_url=PackageURL(
                            type="nuget",
                            namespace="",
                            name=id_,
                            version=version,
                            qualifiers={},
                            subpath="",
                        ).to_string(),
                    )
                )
            except ValidationError as ex:
                LOGGER.warning(
                    "Malformed package. Required fields are missing or data "
                    "types are incorrect.",
                    extra={
                        "extra": {
                            "exception": ex.errors(include_url=False),
                            "location": location.path(),
                        }
                    },
                )
                continue

    return packages, []



from enum import (
    Enum,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from pydantic import (
    BaseModel,
    ConfigDict,
)
from typing import (
    Type,
)


class SourceType(Enum):
    DIRECTORY = "dir"
    DOCKER = "docker"
    DOCKER_DAEMON = "docker-daemon"

    @classmethod
    def from_string(cls: Type["SourceType"], value: str) -> "SourceType":
        for member in cls:
            if member.value == value.lower():
                return member
        raise ValueError(f"{value} is not a valid {cls.__name__}")


class SbomConfig(BaseModel):
    source: str
    source_type: SourceType
    output_format: str
    output: str
    resolver: Resolver | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)



from copy import (
    deepcopy,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.internal.package_information.php import (
    get_composer_package,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.pkg.php import (
    PhpComposerAuthors,
    PhpComposerExternalReference,
    PhpComposerInstalledEntry,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.file import (
    Digest,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)
from typing import (
    Any,
)

LOGGER = logging.getLogger(__name__)


def package_url(name: str, version: str) -> str:
    fields = name.split("/")

    if len(fields) == 1:
        vendor = ""
        name = fields[0]
    elif len(fields) >= 2:
        vendor = fields[0]
        name = "-".join(fields[1:])

    return PackageURL(
        type="composer",
        namespace=vendor,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def new_package_from_composer(
    package: IndexedDict, location: Location, is_dev: bool = False
) -> Package | None:
    new_location = deepcopy(location)
    if new_location.coordinates:
        new_location.coordinates.line = package.position.start.line

    try:
        return Package(
            name=package["name"],
            version=package["version"],
            locations=[new_location],
            language=Language.PHP,
            licenses=list(package.get("license", [])),
            type=PackageType.PhpComposerPkg,
            p_url=package_url(package["name"], package["version"]),
            metadata=PhpComposerInstalledEntry(
                name=package["name"],
                version=package["version"],
                source=PhpComposerExternalReference(
                    type=package["source"]["type"],
                    url=package["source"]["url"],
                    reference=package["source"]["reference"],
                    shasum=package["source"].get("shasum") or None,
                ),
                dist=PhpComposerExternalReference(
                    type=package["dist"]["type"],
                    url=package["dist"]["url"],
                    reference=package["dist"]["reference"],
                    shasum=package["dist"].get("shasum") or None,
                ),
                require=package.get("require"),
                provide=package.get("provide"),
                require_dev=package.get("require-dev"),
                suggest=package.get("suggest"),
                license=package.get("license", []),
                type=package["type"],
                notification_url=package.get("notification-url"),
                bin=package.get("bin", []),
                authors=[
                    PhpComposerAuthors(
                        name=x["name"],
                        email=x.get("email"),
                        homepage=x.get("homepage"),
                    )
                    for x in package.get("authors", [])
                ],
                description=package.get("description"),
                homepage=package.get("homepage"),
                keywords=package.get("keywords", []),
                time=package.get("time"),
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": new_location.path(),
                }
            },
        )
        return None


def package_url_from_pecl(pkg_name: str, version: str) -> str:
    purl = PackageURL(
        type="pecl",
        namespace="",
        name=pkg_name,
        version=version,
        qualifiers=None,
        subpath="",
    )
    return purl.to_string()


def _set_author(package: Package, composer_package: dict[str, Any]) -> Package:
    if not composer_package["authors"]:
        return package
    authors = []
    for author_item in composer_package["authors"]:
        author = author_item["name"]
        if "email" in author_item:
            author += f" <{author_item['email']}>"
        authors.append(author)
    if package.health_metadata and not package.health_metadata.authors:
        package.health_metadata.authors = ", ".join(authors)
    return package


def complete_package(package: Package) -> Package:
    current_package = get_composer_package(package.name, package.version)
    composer_package = get_composer_package(package.name)
    if not composer_package:
        return package

    package.health_metadata = HealthMetadata(
        latest_version=composer_package["version"],
        latest_version_created_at=composer_package["time"],
    )

    if not package.licenses:
        package.licenses = composer_package["license"]

    if current_package:
        digest_value = current_package.get("dist", {}).get("shasum") or None
        package.health_metadata.artifact = Artifact(
            url=current_package["dist"]["url"],
            integrity=Digest(
                algorithm="sha" if digest_value else None,
                value=digest_value,
            ),
        )
    package = _set_author(package, composer_package)

    return package
